'''Evaluators that operate sequentially on ranked votes.

This hosts mainly the transferable vote evaluator
(:class:`TransferableVoteSelector`) but also its less known relative,
:class:`PreferenceAddition`.
'''
import operator
import sys
import itertools
import collections
import logging
from fractions import Fraction
from typing import Any, List, Dict, Tuple, Union, Callable, Optional
from numbers import Number

import votelib.convert
import votelib.persist
import votelib.util
import votelib.component.quota
import votelib.component.transfer
import votelib.evaluate.core
import votelib.evaluate.condorcet
from votelib.candidate import Candidate
from votelib.vote import RankedVoteType
from votelib.persist import simple_serialization
from votelib.component.transfer import VoteTransferer
from votelib.evaluate.condorcet import SeatlessSelector, SmithSet

RankedVoteAllocation = votelib.component.transfer.RankedVoteAllocation

INF = float('inf')

logger = logging.getLogger(__name__)

DEFAULT_TRANSFERER = votelib.component.transfer.Gregory()


@simple_serialization
class TransferableVoteDistributor:
    '''Select candidates by eliminating and transfering votes among them.

    This is the evaluator for transferable vote selection (since the evaluator
    does not concern itself with restrictions on allowed votes, it covers not
    only single transferable vote systems - STV -, but also potential multiple
    transferable vote systems). Its single-winner
    variant is also called instant-runoff voting (IRV).

    First, this evaluator looks at first-preference votes. If any of the
    candidates has at least a specified quota of votes, they are awarded the
    number of seats according to how many times the quota fits into their
    number of votes. If they hit their maximum achievable number of seats
    (if there is one), their votes over the quota are redistributed
    (reallocated) to other candidates according to the next stated preference.

    If no candidate has the quota, the candidate with the fewest
    currently allocated votes is eliminated and their votes transferred
    according to their next stated preference. Votes that have no further
    stated preferences are called exhausted and are removed from consideration.

    The current version does not use elimination breakpoints.

    :param transferer: An instance determining how much votes for eliminated
        candidates to transfer and to whom. It must provide the
        :class:`votelib.component.transfer.VoteTransferer` interface
        (experimental, stability not guaranteed). The basic vote transfer
        variants are implemented in the :mod:`votelib.component.transfer`
        module and can be referred to by their names as strings.
    :param retainer: A selector determining which candidates to retain when
        elimination is to be performed (it may accept a number of seats, which
        will correspond to the number of candidates to retain). If not given,
        the candidates with the lowest amounts of currently allocated votes
        will be eliminated one by one.
    :param eliminate_step: Determines how many candidates to eliminate whenever
        elimination is to be performed. If a negative integer, determines how
        many candidates to eliminate at each step (the default eliminates one
        candidate at a time). If a positive integer, determines how many
        candidates to retain after elimination - this essentially allows only
        a single elimination and might cause an infinite loop if not used
        properly.
    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the
        :mod:`votelib.component.quota` module. If
        None, no election by quota takes place - the candidates are just
        eliminated until the desired number remains. (Not specifying the quota
        only works when the maximum numbers of seats per candidate are
        specified.)
    :param accept_quota_equal: Whether to use non-strict comparison for the
        number of votes against the quota.
    :param mandatory_quota: If True, the candidates must reach the quota
        to be elected, i.e. they cannot be merely the last to be eliminated.
        If False, as soon as the number of unallocated seats equals the
        number of remaining candidates, these are all elected regardless
        of reaching the quota.
    '''
    def __init__(self,
                 transferer: Union[str, VoteTransferer] = DEFAULT_TRANSFERER,
                 retainer: Optional[votelib.evaluate.core.Selector] = None,
                 eliminate_step: Optional[int] = -1,
                 quota_function: Union[
                     str, Callable[[int, int], Number], None
                 ] = 'droop',
                 accept_quota_equal: bool = True,
                 mandatory_quota: bool = False,
                 # use_breakpoints: bool = True, - quota + applied
                 ):
        self.transferer = votelib.component.transfer.construct(transferer)
        if quota_function is not None:
            self.quota_function = votelib.component.quota.construct(
                quota_function
            )
        else:
            self.quota_function = None
        self.accept_quota_equal = accept_quota_equal
        self.mandatory_quota = mandatory_quota
        self.retainer = retainer
        self.eliminate_step = eliminate_step

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Allocate seats to candidates by transferable vote.

        :param votes: Ranked votes. Equal rankings are allowed.
        :param n_seats: Number of seats to allocate to candidates.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the result awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        return self.nth_count(
            votes,
            n_seats,
            sys.maxsize,    # this is not maximum int value but an equivalent
            prev_gains=prev_gains,
            max_seats=max_seats,
        )[1]

    def next_count(self,
                   allocation: RankedVoteAllocation,
                   n_seats: int,
                   total_n_votes: int,
                   prev_gains: Dict[Candidate, int] = {},
                   max_seats: Dict[Candidate, int] = {},
                   ) -> Tuple[RankedVoteAllocation, Dict[Candidate, int]]:
        '''Advance the transferable voting process by one iteration (count).

        :param allocation: Current allocation of ranked votes to candidates
            still contesting the remaining seats. Eliminated candidates are
            not present in the dictionary keys. Exhausted votes are keyed by
            None.
        :param n_seats: Total number of seats to award.
        :param total_n_votes: Total number of votes. Used to compute the
            election quota.
        :param prev_gains: Numbers of seats the candidates gained in the
            previous counts of the election.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        :returns: A 2-tuple containing the new allocation of votes and
            a mapping of candidates to newly assigned seats (might be empty if
            no seats were awarded on this count).
        '''
        n_rem_seats = n_seats - sum(prev_gains.values())
        totals = allocation_totals(allocation)
        avail_seats = {
            cand: max_seats.get(cand, INF) - prev_gains.get(cand, 0)
            for cand in sorted(allocation, key=totals.get, reverse=True)
            if cand is not None
        }
        tot_avail_seats = sum(avail_seats.values())
        if tot_avail_seats == n_rem_seats and not self.mandatory_quota:
            logger.info('electing all remaining: %s', avail_seats)
            return {}, avail_seats    # elect all remaining, no choice
        else:
            logger.info('current vote totals: %s', totals)
            quota_val = self._compute_quota(total_n_votes, n_seats)
            logger.info('quota computed at %g', quota_val)
            quota_elected = self._elect_by_quota(
                totals,
                quota_val,
                n_rem_seats,
                prev_gains=prev_gains,
                max_seats=max_seats
            )
            if quota_elected:
                logger.info('%s elected by quota', quota_elected)
                allocation = self.transferer.subtract(allocation, {
                    cand: n_seats * quota_val
                    for cand, n_seats in quota_elected.items()
                })
                logger.debug('vote totals after election subtraction: %s',
                             allocation_totals(allocation))
                current_seats = votelib.util.sum_dicts(
                    quota_elected, prev_gains
                )
                eliminated = [
                    cand for cand in quota_elected
                    if current_seats.get(cand, 0) >= max_seats.get(cand, INF)
                ]
            else:
                # nobody elected by quota, we have to eliminate
                retained = self.select_retained(totals)
                eliminated = [
                    cand for cand in totals
                    # do not eliminate exhausted votes
                    if cand not in retained and cand is not None
                ]
            if eliminated:
                logger.info('eliminating %s', eliminated)
                new_allocation = self.transferer.transfer(
                    allocation, eliminated
                )
            else:
                new_allocation = allocation
            logger.debug('vote totals after transfer: %s',
                         allocation_totals(new_allocation))
            return new_allocation, quota_elected

    def nth_count(self,
                  votes: Dict[RankedVoteType, Number],
                  n_seats: int = 1,
                  count_number: int = 1,
                  prev_gains: Dict[Candidate, int] = {},
                  max_seats: Dict[Candidate, int] = {},
                  ) -> Tuple[RankedVoteAllocation, Dict[Candidate, int]]:
        '''Get the intermediate counting state at a given iteration (count).

        :param votes: Ranked votes. Equal rankings are allowed.
        :param n_seats: Number of candidates to select.
        :param count_number: 1-indexed count number.
        :returns: A 2-tuple containing the allocation of votes after the given
            count and a list of elected candidates so far (might be empty).
        '''
        allocation = initial_allocation(votes, self.transferer)
        total_n_votes = sum(votes.values())    # needed for quota
        new_allocation = None
        seats = prev_gains.copy()
        for count_i in range(count_number):
            n_already_elected = sum(seats.values())
            if n_already_elected == n_seats:
                logger.info('%d seats allocated, terminating', n_seats)
                break
            if new_allocation is not None:
                allocation = new_allocation
            logger.info('proceeding to count %d', count_i + 1)
            new_allocation, newly_elected = self.next_count(
                allocation,
                n_seats,
                total_n_votes,
                prev_gains=seats,
                max_seats=max_seats,
            )
            if not newly_elected and new_allocation == allocation:
                raise votelib.evaluate.core.VotingSystemError(
                    'infinite loop in STV'
                )
            votelib.util.add_dict_to_dict(seats, newly_elected)
        return allocation_totals(allocation), seats

    def _compute_quota(self,
                       total_n_votes: Optional[Number],
                       n_seats: int
                       ) -> Number:
        if self.quota_function and total_n_votes and n_seats:
            return self.quota_function(total_n_votes, n_seats)
        else:
            return INF

    def _elect_by_quota(self,
                        totals: Dict[Optional[Candidate], Number],
                        quota_val: Number,
                        n_rem_seats: int,
                        prev_gains: Dict[Candidate, int] = {},
                        max_seats: Dict[Candidate, int] = {},
                        ) -> Dict[Candidate, int]:
        quota_multiples = {}
        overcounts = {}
        cand_items = sorted(
            totals.items(),
            key=operator.itemgetter(1),
            reverse=True
        )
        for cand, total in cand_items:
            n_multiples = total // quota_val
            overcount = total - n_multiples * quota_val
            if cand is not None and (self.accept_quota_equal or overcount):
                actual_seats = (
                    min(n_multiples, max_seats.get(cand, INF))
                    - prev_gains.get(cand, 0)
                )
                if actual_seats > 0:
                    quota_multiples[cand] = actual_seats
                    overcounts[cand] = overcount
        if quota_multiples:
            total_awarded = sum(quota_multiples.values())
            if total_awarded > n_rem_seats:
                # Over maximum, we need to select who will get one less seat.
                quota_multiples = self._correct_overcount(
                    quota_multiples, overcounts, n_rem_seats
                )
        return quota_multiples

    def _correct_overcount(self,
                           awarded_seats: Dict[Candidate, int],
                           overcounts: Dict[Candidate, Number],
                           n_rem_seats: int,
                           ) -> Dict[Candidate, int]:
        kept = votelib.evaluate.core.get_n_best(overcounts, n_rem_seats)
        if any(isinstance(e, votelib.evaluate.core.Tie) for e in kept):
            raise NotImplementedError('tie in STV quota election')
        else:
            return {
                cand: (n_seats - 1 if cand not in kept else n_seats)
                for cand, n_seats in awarded_seats.items()
                if cand in kept or n_seats > 1
            }

    def select_retained(self,
                        totals: Dict[Optional[Candidate], Number]
                        ) -> List[Candidate]:
        totals_in_play = {
            cand: tot for cand, tot in totals.items()
            if cand is not None
        }
        if self.retainer:
            if votelib.evaluate.core.accepts_seats(self.retainer):
                n_seats = self._retained_count(totals_in_play)
                retained = self.retainer.evaluate(totals_in_play, n_seats)
            else:
                retained = self.retainer.evaluate(totals_in_play)
        else:
            n_seats = self._retained_count(totals_in_play)
            retained = votelib.evaluate.core.get_n_best(totals, n_seats)
        if any(isinstance(e, votelib.evaluate.core.Tie) for e in retained):
            raise NotImplementedError('tie in STV elimination')
        return retained

    def _retained_count(self, totals: Dict[Candidate, Number]) -> int:
        if self.eliminate_step is None:
            raise ValueError('need to specify eliminate step'
                             ' without standalone retainer')
        if self.eliminate_step < 0:
            return max(len(totals) + self.eliminate_step, 1)
        else:
            return min(self.eliminate_step, len(totals) - 1)


def initial_allocation(votes: Dict[RankedVoteType, Number],
                       transferer: VoteTransferer = DEFAULT_TRANSFERER,
                       ) -> RankedVoteAllocation:
    '''Allocate votes by first preference.

    :param votes: Ranked votes.
    :param transferer: A :class:`votelib.component.transfer.VoteTransferer`
        to transfer votes from shared first ranks among their candidates.
    :returns: The votes dictionary separated into subdictionaries keyed by
        candidate to whom the votes are allocated. A candidate with no
        first preference votes will be assigned to an empty dictionary.
    '''
    first_prefs = {
        cand: {} for cand in votelib.util.all_ranked_candidates(votes)
    }
    FICTIONAL = object()
    for vote, n_votes in votes.items():
        if not vote:
            continue    # for empty votes
        first_pref = vote[0]
        if isinstance(first_pref, collections.abc.Set):
            # first rank is shared
            # we pass this to transferer - we prepend a fictional (None)
            # eliminated candidate to this vote and have it transferred
            if FICTIONAL not in first_prefs:
                first_prefs[FICTIONAL] = {}
            first_prefs[FICTIONAL][(FICTIONAL,) + vote] = n_votes
        else:
            first_prefs[first_pref][vote] = n_votes
    if first_prefs.get(FICTIONAL):
        split_alloc = transferer.transfer(
            first_prefs, candidates=[FICTIONAL],
        )
        for cand_alloc in split_alloc.values():
            to_remove = []
            to_update = {}
            for vote, n_votes in cand_alloc.items():
                if vote[0] is FICTIONAL:
                    to_remove.append(vote)
                    to_update[vote[1:]] = n_votes
            for vote in to_remove:
                del cand_alloc[vote]
            cand_alloc.update(to_update)
        return split_alloc
    else:
        return first_prefs


def allocation_totals(allocation: RankedVoteAllocation
                      ) -> Dict[Candidate, Number]:
    return {
        cand: sum(cand_votes.values())
        for cand, cand_votes in allocation.items()
    }


class TransferableVoteSelector:
    def __init__(self, *args, **kwargs):
        got_inner_pos_arg = (
            len(args) == 1 and not kwargs
            and isinstance(args[0], TransferableVoteDistributor)
        )
        if got_inner_pos_arg:
            self._inner = args[0]
        elif '_inner' in kwargs:
            self._inner = kwargs['_inner']
        else:
            self._inner = TransferableVoteDistributor(*args, **kwargs)

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        return self.nth_count(votes, n_seats, sys.maxsize)[1]

    def next_count(self,
                   allocation: RankedVoteAllocation,
                   n_seats: int,
                   total_n_votes: int,
                   elected: List[Candidate] = [],
                   ) -> Tuple[RankedVoteAllocation, List[Candidate]]:
        '''Advance the transferable voting process by one iteration (count).

        :param allocation: Current allocation of ranked votes to candidates
            still contesting the remaining seats.
        :param n_seats: Total number of seats to award.
        :param total_n_votes: The total number of votes cast in the election.
            Used to determine the election quota, if enabled.
        :param elected: Candidates already elected (to be skipped when
            transferring votes).
        :returns: A 2-tuple containing the new allocation of votes and
            a mapping of candidates to newly assigned seats (might be empty if
            no seats were awarded on this count).
        '''
        all_cands = set()
        for cand, alloc_votes in allocation:
            all_cands.update(votelib.util.all_ranked_candidates(alloc_votes))
        new_alloc, newly_elected = self._inner.next_count(
            allocation,
            n_seats,
            total_n_votes,
            prev_gains={c: 1 for c in elected},
            max_seats={c: 1 for c in all_cands}
        )
        return new_alloc, votelib.util.distribution_to_selection(newly_elected)

    def nth_count(self,
                  votes: Dict[RankedVoteType, Number],
                  n_seats: int = 1,
                  count_number: int = 1,
                  ) -> Tuple[RankedVoteAllocation, List[Candidate]]:
        '''Get the intermediate counting state at a given iteration (count).

        :param votes: Ranked votes. Equal rankings are allowed.
        :param n_seats: Number of candidates to select.
        :param count_number: 1-indexed count number.
        :returns: A 2-tuple containing the allocation of votes after the given
            count and a list of elected candidates so far (might be empty).
        '''
        all_cands = votelib.util.all_ranked_candidates(votes)
        allocation, elected_dict = self._inner.nth_count(
            votes,
            n_seats,
            count_number,
            max_seats={c: 1 for c in all_cands}
        )
        return allocation, votelib.util.distribution_to_selection(elected_dict)

    @property
    def quota_function(self):
        return self._inner.quota_function

    def to_dict(self) -> Dict[str, Any]:
        return {
            'class': votelib.persist.scoped_class_name(self),
            '_inner': self._inner.to_dict(),
        }


# Gradual preference addition
@simple_serialization
class PreferenceAddition:
    '''Evaluates ranked votes by stepwise addition of lower preferences.

    Each candidate starts with their first-preference votes and
    lower-preference votes are added to them until a sufficient amount of
    candidates achieve a majority. This includes the following voting systems:

    -   *Bucklin voting* where the lower preferences are added without change.
    -   *Oklahoma system* where the lower preferences are added divided by
        the order of preference, thus yielding the row of coefficients
        1, 1/2, 1/3, 1/4... (Use :class:`Fraction` to maintain
        exact values.)

    :param coefficients: Coefficients to multiply the lower preferences with
        before they are added to the vote totals (the default adds them simply,
        which creates Bucklin voting).
    :param split_equal_rankings: Whether to split votes having multiple
        alternatives at the same rank, or to add the whole amount to each
        of these alternatives.
    '''
    def __init__(self,
                 coefficients: Union[
                     List[Number],
                     Callable[[int], Number],
                 ] = [1],
                 split_equal_rankings: bool = True,
                 ):
        self.coefficients = coefficients
        self.split_equal_rankings = split_equal_rankings

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by sequential preference addition.

        :param votes: Ranked votes.
        :param n_seats: Number of candidates to be elected.
        '''
        if self.split_equal_rankings:
            votes = self._decouple_equal_rankings(votes)
        total_votes = collections.defaultdict(int)
        majority_quota = Fraction(sum(votes.values()), 2)
        max_pref_len = max(
            len(preferences) for preferences, n_votes in votes.items()
        )
        elected = []
        for pref_i in range(max_pref_len):
            # add this round's preferences
            self._add_round_votes(total_votes, votes, pref_i, elected)
            # take all that have achieved majority, ordered by the vote sum
            majority = {
                cand: n_votes
                for cand, n_votes in votelib.util.sorted_votes(total_votes)
                if n_votes > majority_quota
            }
            best = votelib.evaluate.core.get_n_best(
                majority,
                n_seats - len(elected)
            )
            elected.extend(best)
            if len(elected) == n_seats:
                break
            else:
                for cand in best:
                    if cand in total_votes:
                        del total_votes[cand]
            # if still not enough elected, go on by adding another preference
        return votelib.evaluate.core.Tie.reconcile(elected)

    def _decouple_equal_rankings(self,
                                 votes: Dict[RankedVoteType, int]
                                 ) -> Dict[RankedVoteType, Fraction]:
        new_votes = None
        for one_vote, n_votes in votes.items():
            equal_rank_tuples = [
                (i, rank) for i, rank in enumerate(one_vote)
                if isinstance(rank, collections.abc.Set)
            ]
            if equal_rank_tuples:
                if new_votes is None:
                    new_votes = votes.copy()
                del new_votes[one_vote]
                eqr_is = [i for i, rank in equal_rank_tuples]
                variants = list(itertools.product(*(
                    itertools.permutations(rank)
                    for i, rank in equal_rank_tuples
                )))
                n_variant_votes = Fraction(n_votes, len(variants))
                for variant in variants:
                    var_vote = one_vote[:]
                    offset = 0
                    for i, var_part in zip(eqr_is, variant):
                        var_vote = var_vote[:]
                        var_vote = (
                            var_vote[:i+offset]
                            + var_part
                            + var_vote[i+offset+1:]
                        )
                        offset += len(var_part)
                    new_votes[var_vote] = n_variant_votes
        return new_votes if new_votes is not None else votes

    def _add_round_votes(self,
                         total_votes: Dict[Candidate, Fraction],
                         votes: Dict[RankedVoteType, int],
                         pref_i: int,
                         elected: List[Candidate],
                         ) -> None:
        coef = self._get_coefficient(pref_i)
        for preferences, n_votes in votes.items():
            if pref_i < len(preferences):
                preference = preferences[pref_i]
                if isinstance(preference, collections.abc.Set):
                    n_to_add = n_votes * coef
                    for cand in preference:
                        if cand not in elected:
                            total_votes[cand] += n_to_add
                elif preference not in elected:
                    total_votes[preference] += n_votes * coef

    def _get_coefficient(self, pref_i: int) -> Number:
        if hasattr(self.coefficients, '__call__'):
            return self.coefficients(pref_i)
        elif pref_i < len(self.coefficients):
            return self.coefficients[pref_i]
        else:
            return self.coefficients[-1]


RANKED_SUBSETTER = votelib.convert.SubsettedVotes(
    votelib.vote.RankedSubsetter()
)
RANKED_TO_CONDORCET = votelib.convert.RankedToCondorcetVotes()


class TidemanAlternative:
    '''Tideman alternative selector (Alternative Smith/Schwartz). [#tidaltw]_

    Uses a Smith or Schwartz set selector; if the set has multiple members,
    eliminates one candidate by transferable vote (instant-runoff) and reruns.
    For election of multiple candidates, it runs in multiple tiers where
    previous winners are eliminated from all following tiers. [#gacondo]_

    :param set_selector: The Smith/Schwartz set selector. Any seatless selector
        that accepts the Condorcet (pairwise preference) vote format.

    .. [#tidaltw] "Tideman alternative method", Wikipedia.
        https://en.wikipedia.org/wiki/Tideman_alternative_method
    .. [#gacondo] Green-Armytage, James. "Four Condorcet-Hare Hybrid Methods
        for Single-Winner Elections", Voting Matters.
        http://www.votingmatters.org.uk/ISSUE29/I29P1.pdf
    '''
    def __init__(self,
                 set_selector: SeatlessSelector = SmithSet(),
                 ):
        self.set_selector = set_selector

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        ranked_set = []
        eligible_set = set(votelib.util.all_ranked_candidates(votes))
        tier_votes = votes
        while True:
            winner = self.run_tier(tier_votes)
            logger.info('adding %s to winners', winner)
            ranked_set.append(winner)
            eligible_set.remove(winner)
            if len(ranked_set) == n_seats or not eligible_set:
                return ranked_set
            else:
                tier_votes = RANKED_SUBSETTER.convert(tier_votes)

    def run_tier(self, votes: Dict[RankedVoteType, int]) -> Candidate:
        round_votes = votes
        while round_votes:
            s_set_list = self.get_winner_set(round_votes)
            logger.info('condorcet set: %s', s_set_list)
            if len(s_set_list) == 1:
                return s_set_list.pop()
            else:
                round_votes = RANKED_SUBSETTER.convert(round_votes, s_set_list)
                rem = eliminate_one(round_votes)
                logger.info('eliminated to %s', rem)
                if len(rem) == 1:
                    return rem.pop()
                else:
                    round_votes = RANKED_SUBSETTER.convert(round_votes, rem)
        raise NotImplementedError

    def get_winner_set(self,
                       votes: Dict[RankedVoteType, int],
                       ) -> List[Candidate]:
        return self.set_selector.evaluate(RANKED_TO_CONDORCET.convert(votes))


def eliminate_one(votes: Dict[RankedVoteType, int]) -> List[Candidate]:
    return votelib.evaluate.core.get_n_best(
        allocation_totals(initial_allocation(votes)),
        len(votelib.util.all_ranked_candidates(votes)) - 1
    )


class Benham:
    '''Benham sequential Condorcet selector. [#gacondo]_

    Selects a Condorcet winner. If one does not exist, eliminates one candidate
    using transferable vote (instant-runoff) and re-runs.
    '''
    CONDO = votelib.evaluate.condorcet.CondorcetWinner()

    def evaluate(self,
                 votes: Dict[RankedVoteType, int],
                 n_seats: int = 1) -> List[Candidate]:
        assert n_seats == 1
        current_votes = votes
        condowin = self.get_condorcet_winner(current_votes)
        while condowin is None:
            remains = eliminate_one(current_votes)
            if len(remains) == 1:
                return remains
            else:
                current_votes = RANKED_SUBSETTER.convert(votes, remains)
                condowin = self.get_condorcet_winner(current_votes)
        return [condowin]

    def get_condorcet_winner(self,
                             votes: Dict[RankedVoteType, int],
                             ) -> Optional[Candidate]:
        condores = self.CONDO.evaluate(RANKED_TO_CONDORCET.convert(votes))
        return condores[0] if condores else None



"""Generate votes for voting system simulations.

The generators implemented here mostly produce score votes since those are
the most general type. For other types, convert the result using converters:

-   To get ranked votes from score votes, use
    :class:`votelib.convert.ScoreToRankedVotes`.
-   To get simple votes from score votes, use
    :class:`votelib.convert.ScoreToRankedVotes`
    and :class:`votelib.convert.RankedToFirstPreference`
    (wrapped in :class:`votelib.convert.Chain`).
-   To get approval votes over a threshold, use
    :class:`votelib.convert.ScoreToApprovalVotesThreshold`.
"""

import sys
import abc
import math
import string
import operator
import collections
import random
from numbers import Number
from typing import Any, Optional, List, Tuple, Dict, Union, Iterable

import votelib.vote
import votelib.convert
from votelib.candidate import Candidate


class Sampler(metaclass=abc.ABCMeta):
    """A generic sampler interface."""
    n_dims: int

    def sample(self,
               n: int,
               n_dims: Optional[int] = None,
               ) -> Iterable[Tuple[float, ...]]:
        raise NotImplementedError


class DistributionSampler(Sampler):
    """Sample points from the issue or score space by specifying a distribution.

    Uses a statistical probability distribution to produce randomly located
    points within the issue space or score space of given dimensionality.
    The distributions are taken from Python's *random* module by referencing
    the names of the generating functions.

    The outputs from this sampler need to be fed into
    :class:`IssueSpaceGenerator` to produce votes or specify candidate
    positions or into :class:`ScoreSpaceGenerator` to be converted to candidate
    scorings directly.

    Any superfluous keyword arguments are passed to the generating function
    from the random module. If no keyword arguments are given but are required
    (i.e. the distribution parameters need to be specified), for some
    distributions (uniform, gauss, triangular, beta), defaults are specified
    in this class and automatically used if necessary.

    :param distribution: The name of the distribution to use. Must refer to a
        name of a function in Python stdlib random module that produces random
        floats.
    :param n_dims: Dimensionality of the issue or score space to sample from.
    """
    DEFAULT_PARAMS: Dict[str, Dict[str, Union[Number, Tuple[Number, ...]]]] = {
        'gauss': {'mu': 0, 'sigma': 1},
        'uniform': {'a': 0, 'b': 1},
        'triangular': {'low': 0, 'high': 1, 'mode': .5},
        'beta': {'alpha': 2, 'beta': 2},
    }

    def __init__(self,
                 distribution: str = 'gauss',
                 n_dims: Optional[int] = None,
                 **kwargs):
        self.distro_fx = getattr(random, distribution)
        if not kwargs and distribution in self.DEFAULT_PARAMS:
            kwargs = self.DEFAULT_PARAMS[distribution].copy()
        for argname, argval in kwargs.items():
            if hasattr(argval, '__len__') and not isinstance(argval, str):
                if n_dims is None:
                    n_dims = len(argval)
                elif len(argval) != n_dims:
                    raise ValueError(
                        f'sampling {argname} parameter has'
                        f'{len(argval)} dimensions, expected {n_dims}'
                    )
            elif isinstance(argval, Number) and n_dims is not None:
                kwargs[argname] = tuple([argval] * n_dims)
        self.n_dims = n_dims
        if self.n_dims is None:
            self.gener_args = kwargs
        else:
            self.gener_args = tuple(
                {argname: argval[i] for argname, argval in kwargs.items()}
                for i in range(self.n_dims)
            )

    def sample(self,
               n: int,
               n_dims: Optional[int] = None,
               ) -> Iterable[Tuple[float, ...]]:
        """Sample n issue space samples from the distribution."""
        if self.n_dims is None:
            if n_dims is None:
                raise ValueError('need n_dims arg when not set on instance')
            else:
                gener_args = tuple([self.gener_args] * n_dims)
        elif n_dims is not None and n_dims != self.n_dims:
            raise ValueError(f'conflicting n_dims: got {n_dims}'
                             f'but {self.n_dims} set on instance')
        else:
            gener_args = self.gener_args
        for i in range(n):
            yield tuple(self.distro_fx(**kwargs) for kwargs in gener_args)


class BoundedSampler(Sampler):
    """A sampler from a bounded issue space.

    Wraps another sampler to only produce issue space samples that lie within
    a specified multidimensional bounding box. Useful e.g. for the generation
    of Yee diagrams.

    The outputs from this sampler need to be fed into
    :class:`SamplingGenerator` to produce votes or specify candidate positions.

    :param inner: A sampler (e.g. :class:`DistributionSampler`) to wrap.
        Its samples are filtered by the specified bounding box.
    :param bbox: The bounding box to restrict the samples to. First, all
        minima per dimension are specified, then all maxima; for two
        dimensions, this would be ``(minx, miny, maxx, maxy)``. If the inner
        sampler has a defined number of dimensions, it is also possible
        to specify only two numbers, which will be interpreted as the minimum
        and maximum in each dimension.
    """
    def __init__(self, inner: Sampler, bbox: Tuple[Number, ...]):
        self.inner = inner
        if len(bbox) % 2 != 0:
            raise ValueError('bounding box must have an odd number of'
                             f'coordinates, got {len(bbox)}')
        if len(bbox) == 2 and self.inner.n_dims is not None:
            bbox = tuple(
                [bbox[0]] * self.inner.n_dims
                + [bbox[1]] * self.inner.n_dims
            )
        self.bbox = bbox
        self._mins = self.bbox[:len(bbox) // 2]
        self._maxs = self.bbox[len(bbox) // 2:]

    def sample(self, n: int, *args, **kwargs) -> Iterable[Tuple[float, ...]]:
        """A generator to sample n bbox-restricted issue space samples."""
        n_yielded = 0
        for coors in self.inner.sample(sys.maxsize, *args, **kwargs):
            is_in_bbox = (
                all(c >= m for c, m in zip(coors, self._mins))
                and all(c <= m for c, m in zip(coors, self._maxs))
            )
            if is_in_bbox:
                yield coors
                n_yielded += 1
                if n_yielded == n:
                    break


class Generator:
    def generate(self, n: int) -> Dict[Any, int]:
        raise NotImplementedError


class IssueSpaceGenerator:
    """Generate random votes by sampling from a multidimensional issue space.

    The most common paradigm for random vote generation is to spread some
    candidates as points into a multidimensional space (representing their
    opinions or characteristics), then sample voter points from a statistical
    distribution on that space and determine their votes based on their
    proximity to candidates. This class implements this paradigm, delegating
    the sampling to a contained instance of :class:`Sampler`.

    :param candidates: Positions of candidates in the issue space.
        It is also possible to specify just an integer; in that case, the
        candidate positions will be sampled in the same way the voters are.
    :param sampler: How to sample the voter points. Either an object with a
        ``sample()`` method that yields numerical tuples with the correct
        number of dimensions (such as instances of :class:`Sampler`
        subclasses), or a string referencing a name of a statistical
        distribution; in that case, :class:`DistributionSampler` will be
        invoked in two dimensions with default settings on the specified
        distribution. See the class documentation for more on supported
        distributions.
    :param vote_creation: How to transform voter-to-candidate proximities
        in the issue space to score votes:

        -   ``minmax``: The closest candidate is assigned a score of 1, the
            furthest is assigned a score of 0. Score votes are produced.
        -   ``closest``: The closest candidate is voted for. Simple votes are
            produced.
        -   Other transformations are not implemented.

    :param random_state: Seed for the sampler.
    """
    def __init__(self,
                 candidates: Union[int, Dict[Candidate, Tuple[float, ...]]],
                 sampler: Union[str, Sampler] = 'gauss',
                 vote_creation: str = 'minmax',
                 random_state: Optional[int] = None,
                 ):
        if not hasattr(self, vote_creation):
            raise ValueError(f'invalid vote creation method: {vote_creation}')
        self.sampler = _create_sampler(sampler, n_dims=2)
        self.vote_creation = vote_creation
        self.candidates = candidates
        self.random_state = random_state

    def generate(self, n: int) -> Dict[Any, int]:
        """Generate n votes for the candidate setup."""
        if self.random_state is not None:
            random.seed(self.random_state)
        if isinstance(self.candidates, int):
            candidates = self._create_candidates(n=self.candidates)
            n_dims = None
        else:
            candidates = self.candidates
            n_dims = len(next(iter(candidates.values())))

        return self.samples_to_votes(
            self.sampler.sample(n, n_dims=n_dims),
            candidates
        )

    def _create_candidates(self, n: int) -> Dict[Candidate, Tuple[float, ...]]:
        return dict(zip(
            candidate_names(n),
            self.sampler.sample(n=self.candidates)
        ))

    def samples_to_votes(self,
                         sample: Iterable[Tuple[float, ...]],
                         candidates: Dict[Candidate, Tuple[float, ...]],
                         ) -> Dict[Any, int]:
        """Convert issue space samples to votes."""
        vote_create_fx = getattr(self, self.vote_creation)
        votes = collections.defaultdict(int)
        cands = list(candidates.keys())
        cand_coors = list(candidates.values())
        for vote_coor in sample:
            dists = [
                math.hypot(*[v - c for v, c in zip(vote_coor, cand_coor)])
                for cand_coor in cand_coors
            ]
            votes[vote_create_fx(dists, cands)] += 1
        return dict(votes)

    @staticmethod
    def minmax(distances: List[float],
               candidates: List[Candidate],
               ) -> votelib.vote.ScoreVoteType:
        mindist, maxdist = min(distances), max(distances)
        distrange = maxdist - mindist
        return frozenset(zip(
            candidates,
            [(maxdist - dist) / distrange for dist in distances]
        ))

    @staticmethod
    def closest(distances: List[float],
                candidates: List[Candidate],
                ) -> Candidate:
        return min(zip(distances, candidates), key=operator.itemgetter(0))[1]


class ScoreSpaceGenerator:
    """Randomly sample independent candidate scorings.

    This generator is simpler than :class:`IssueSpaceGenerator` since it
    uses the underlying sampler to produce candidate scorings directly.
    The candidate scores are independent of each other.

    Under default settings, this generator produces the *Impartial Culture*
    (IC). By providing different settings for different dimensions
    of the passed sampler, score probabilities for individual candidates
    can be set. An example for two candidates, in which A will, on average,
    get higher scores (mean 0.7) than B (mean 0.3)::

        ScoreSpaceGenerator(
            candidates=['A', 'B'],
            sampler=BoundedSampler(
                DistributionSampler('gauss', mu=(0.7, 0.3), sigma=(1, 1)),
                bbox=(0, 1)
            )
        )

    :param candidates: Candidate objects (most straightforwardly, a list of
        candidate names as strings). It is also possible to specify just an
        integer; in that case, the candidate names are autogenerated as
        uppercase ASCII letters (A, B, C...).
    :param sampler: How to sample the scores. Either an object with a
        ``sample()`` method that yields numerical tuples with the correct
        number of dimensions (such as instances of :class:`Sampler`
        subclasses), or a string referencing a name of a statistical
        distribution; in that case, :class:`DistributionSampler` will be
        invoked with default settings on the specified distribution.
        See the class documentation for more on supported distributions.
    :param round_scores: Whether to round the scores (using round half to even)
        to integers.
    :param random_state: Seed for the sampler.
    """

    def __init__(self,
                 candidates: Union[int, List[Candidate]],
                 sampler: Union[str, Sampler] = 'uniform',
                 round_scores: bool = False,
                 random_state: Optional[int] = None,
                 ):
        if isinstance(candidates, int):
            if candidates > 26:
                raise NotImplementedError
            candidates = list(string.ascii_uppercase[:candidates])
        self.candidates = candidates
        self.sampler = _create_sampler(sampler, n_dims=None)
        self.round_scores = round_scores
        self.random_state = random_state

    def generate(self, n: int) -> Dict[Any, int]:
        """Generate n votes."""
        if self.random_state is not None:
            random.seed(self.random_state)
        n_cands = len(self.candidates)
        votes = collections.defaultdict(int)
        for scoring in self.sampler.sample(n, n_dims=n_cands):
            if self.round_scores:
                scoring = tuple(int(s) for s in scoring)
            votes[frozenset(zip(self.candidates, scoring))] += 1
        return dict(votes)


def _create_sampler(sampler: Union[str, Sampler], n_dims: Union[int, None]):
    if hasattr(sampler, 'sample'):
        return sampler
    else:
        return DistributionSampler(distribution=sampler, n_dims=n_dims)


def candidate_names(n: int) -> List[str]:
    if n > 26:
        raise NotImplementedError
    return list(string.ascii_uppercase[:n])



'''Condorcet selection evaluators.

These evaluators work by examining pairwise orderings between candidates
(how many voters prefer one candidate to another), which is also the form of
votes they take in; if you have ranked vote input, use
:class:`votelib.convert.RankedToCondorcetVotes` to convert it first and
account for possible shared rankings and unranked candidates.

All of the methods in this module reliably select a Condorcet winner when there
is one in the input.

These evaluators are considered more advanced in terms of satisfied criteria
than transferable vote systems but still cannot defy some impossibility
theorems such as Arrow's or Gibbard's.

These evaluators only take few parameters; therefore, a dictionary of their
instances with different setups is provided in the ``EVALUATORS`` module
variable.
'''

import itertools
import collections
from typing import List, Tuple, Dict, Union, Callable, Collection
from numbers import Number

import votelib.evaluate.core
import votelib.component.pairwin_scorer
from votelib.candidate import Candidate
from votelib.persist import simple_serialization


def pairwise_wins(votes: Dict[Tuple[Candidate, Candidate], int],
                  include_ties: bool = False,
                  ) -> List[Tuple[Candidate, Candidate]]:
    """Select pairs of candidates where the first is preferred to the second.

    :param votes: Condorcet votes (counts of candidate pairs as they appear
        in the voter rankings); use
        :class:`votelib.convert.RankedToCondorcetVotes`
        to produce them from ranked votes.
    :param include_ties: Whether to include pairs of candidates that are tied.
        Such a pair will be included in both directions.
    :returns: Ordered pairs from the input that are generally preferred to the
        opposite ranking (i.e. listed in this order by more voters).
    """
    wins = []
    for pair, count in votes.items():
        upper_cand, lower_cand = pair
        anti_count = votes.get((lower_cand, upper_cand), 0)
        if anti_count < count or include_ties and anti_count == count:
            wins.append(pair)
    return wins


def beat_counts(votes: Dict[Tuple[Candidate, Candidate], int]
                ) -> Dict[Candidate, int]:
    """Count the number of candidates a given candidate beats pairwise.

    :param votes: Condorcet votes (counts of candidate pairs as they appear
        in the voter rankings); use
        :class:`votelib.convert.RankedToCondorcetVotes`
        to produce them from ranked votes.
    """
    n_beats = collections.defaultdict(int)
    for winner, loser in pairwise_wins(votes):
        n_beats[winner] += 1
    return dict(n_beats)


def _smith_schwartz_set(votes: Dict[Tuple[Candidate, Candidate], int],
                        ties: bool = True,
                        ) -> List[Candidate]:
    wins = pairwise_wins(votes, include_ties=ties)
    copeland_scores = Copeland.scores(wins)
    copeland_ordering = list(sorted(
        copeland_scores,
        key=copeland_scores.get,
        reverse=True,
    ))
    cand_copes = dict(zip(copeland_ordering, range(len(copeland_ordering))))
    end_i = 1  # index of first candidate out of smith set
    # Sort wins so that wins over the most promising candidates go first.
    wins.sort(key=lambda tup: copeland_ordering.index(tup[1]))
    for winner, loser in wins:
        if cand_copes[winner] >= end_i and cand_copes[loser] < end_i:
            end_i = cand_copes[winner] + 1
            if end_i == len(copeland_ordering):
                break
    return copeland_ordering[:end_i]


class Selector:
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        raise NotImplementedError


class SeatlessSelector:
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 ) -> List[Candidate]:
        raise NotImplementedError


@simple_serialization
class CondorcetWinner:
    """Condorcet winner selector.

    Selects a candidate that pairwise beats all other candidates, if there
    is one, or an empty list otherwise.
    """
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 ) -> List[Candidate]:
        """Select the Condorcet winner.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        """
        n_required_wins = len(
            frozenset(cand for pair in votes for cand in pair)
        ) - 1
        for cand, n_beats in beat_counts(votes).items():
            if n_beats == n_required_wins:
                return [cand]
        return []


@simple_serialization
class SmithSet:
    """Smith set selector.

    The Smith set is the smallest possible non-empty set whose candidates beat
    all other candidates in pairwise preference comparisons.
    """
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 ) -> List[Candidate]:
        """Select the Smith set.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        """
        return _smith_schwartz_set(votes, ties=True)


@simple_serialization
class SchwartzSet:
    """Schwartz set selector.

    The Schwartz set is the smallest possible non-empty set whose candidates
    are pairwise unbeaten by all other candidates.
    """
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 ) -> List[Candidate]:
        """Select the Schwartz set.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        """
        return _smith_schwartz_set(votes, ties=False)


@simple_serialization
class Copeland:
    '''Copeland (count of pairwise wins) Condorcet selection evaluator.

    Calculates the pairwise wins, constructs the Copeland score by taking
    ``number_of_pairwise_wins - number_of_pairwise_losses`` for each candidate,
    and uses this score to rank candidates. This often produces ties, which
    can be used by second order tiebreaking - by preferring the candidates
    who have pairwise beaten the candidates with the highest total Copeland
    score.

    :param second_order: Whether to use second-order Copeland tiebreaking.
    '''
    def __init__(self, second_order: bool = True):
        self.second_order = second_order

    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates using the Copeland method.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        :param n_seats: Number of candidates to select.
        '''
        wins = pairwise_wins(votes)
        scores = self.scores(wins)
        best = votelib.evaluate.core.get_n_best(scores, n_seats)
        if self.second_order and votelib.evaluate.core.Tie.any(best):
            return self.break_second_order(best, scores, wins)
        else:
            return best

    @staticmethod
    def scores(wins: List[Tuple[Candidate, Candidate]]
               ) -> Dict[Candidate, int]:
        scores = collections.defaultdict(int)
        for winner, loser in wins:
            scores[winner] += 1
            scores[loser] -= 1
        return dict(scores)

    def break_second_order(self,
                           best: List[Union[
                               Candidate,
                               votelib.evaluate.core.Tie
                           ]],
                           scores: Dict[Candidate, int],
                           wins: List[Tuple[Candidate, Candidate]],
                           ) -> List[Candidate]:
        untied = []
        tied = set()
        for selected in best:
            if isinstance(selected, votelib.evaluate.core.Tie):
                tied.update(selected)
            else:
                untied.append(selected)
        second_order_scores = collections.defaultdict(int)
        for winner, loser in wins:
            if winner in tied:
                second_order_scores[winner] += scores[loser]
        return untied + votelib.evaluate.core.get_n_best(
            second_order_scores, len(best) - len(untied)
        )


@simple_serialization
class Schulze:
    '''Schulze (beatpath) Condorcet selection evaluator.

    Also called Schwartz Sequential dropping or path voting. Finds paths
    between pairs of candidates in which each candidate pairwise beats the next
    and then selects the candidates with strongest such paths.
    '''
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates using the Schulze method.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        :param n_seats: Number of candidates to select.
        '''
        paths = self.widest_paths(votes)
        scores = collections.defaultdict(int)
        for winner, loser in pairwise_wins(paths):
            scores[winner] += 1
            scores[loser]    # to make zero appear if not present
        return votelib.evaluate.core.get_n_best(scores, n_seats)

    @staticmethod
    def widest_paths(counts: Dict[Tuple[Candidate, Candidate], Number]
                     ) -> Dict[Tuple[Candidate, Candidate], Number]:
        paths = {}
        all_candidates = set()
        for pair, count in counts.items():
            all_candidates.update(pair)
            if counts.get(tuple(reversed(pair)), 0) < count:
                paths[pair] = count
        all_candidates = list(all_candidates)
        for cand1 in all_candidates:
            for cand2 in all_candidates:
                if cand1 != cand2:
                    for cand_aug in all_candidates:
                        if cand_aug not in (cand1, cand2):
                            paths[cand2, cand_aug] = max(
                                paths.get((cand2, cand_aug), 0),
                                min(
                                    paths.get((cand2, cand1), 0),
                                    paths.get((cand1, cand_aug), 0),
                                )
                            )
        return paths


@simple_serialization
class KemenyYoung:
    '''Kemeny-Young Condorcet selection evaluator.

    Kemeny-Young orders the candidates based on their pairwise comparison by
    constructing an objective function that measures the quality of any
    ordering, evaluating it for all permutations of the candidate set, and
    selecting the ordering with the maximum score.

    The objective function is given as the number of satisfied pairwise
    orderings of candidates as given by the voters.

    WARNING: Due to the enumeration of all candidate set permutations, this
    method is highly computationally expensive (``O(n!)`` in the number of
    candidates) and infeasible on common machines for more than a handful of
    candidates.
    '''
    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by the Kemeny-Young Condorcet method.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        :param n_seats: Number of candidates to select.
        '''
        all_candidates = frozenset(cand for pair in votes for cand in pair)
        best_variants = []
        best_score = 0
        for variant in itertools.permutations(all_candidates):
            score = self.score(variant, votes)
            if score >= best_score:
                variant = list(variant)
                if score > best_score:
                    best_variants = [variant]
                    best_score = score
                else:
                    best_variants.append(variant)
        if len(best_variants) == 1:
            return best_variants[0][:n_seats]
        else:
            return votelib.evaluate.core.Tie.tie_rankings(
                best_variants
            )[:n_seats]

    @staticmethod
    def score(variant: Collection[Candidate],
              votes: Dict[Tuple[Candidate, Candidate], int],
              ) -> int:
        '''Compute the Kemeny-Young ordering score (objective function value).

        :param variant: The ordering of candidates to evaluate.
        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        '''
        return sum(
            votes.get((cand_upper, cand_lower), 0)
            for i, cand_upper in enumerate(variant)
                for cand_lower in variant[i+1:]    # noqa: E131
        )


@simple_serialization
class MinimaxCondorcet:
    '''Minimax Condorcet selection evaluator.

    Also known as successive reversal or Simpson-Kramer method.
    Selects as the winner the candidate whose greatest pairwise defeat is
    smaller than the greatest pairwise defeat of any other candidate.

    The magnitude of the pairwise defeat can be measured in different ways
    according to the pairwise win scorer provided.

    :param pairwin_scoring: A pairwise win scorer callable. Most common
        variants are found in the :mod:`votelib.component.pairwin_scorer`
        module and can be referred to by their names.
    '''
    def __init__(self,
                 pairwin_scoring: Union[str, Callable] = 'winning_votes',
                 ):
        self.pairwin_scoring = votelib.component.pairwin_scorer.construct(
            pairwin_scoring
        )

    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by the Minimax Condorcet method.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        :param n_seats: Number of candidates to select.
        '''
        max_counterscore = {}
        for pair, score in self.pairwin_scoring(votes).items():
            max_counterscore[pair[1]] = max(
                max_counterscore.get(pair[1], -float('inf')),
                score
            )
        return votelib.evaluate.core.get_n_best(
            {cand: -score for cand, score in max_counterscore.items()},
            n_seats
        )


@simple_serialization
class RankedPairs:
    '''Tideman's ranked pairs Condorcet selection evaluator.

    Ranks pairwise wins by their magnitude and sequentially locks pairs of
    who beats whom in descending order into a ranking, discarding pairs that
    would contradict previously established rankings (i.e. create a cycle).

    The magnitude of the pairwise win can be measured in different ways
    according to the pairwise win scorer provided.

    :param pairwin_scoring: A pairwise win scorer callable. Most common
        variants are found in the :mod:`votelib.component.pairwin_scorer`
        module and can be referred to by their names.
    '''
    def __init__(self,
                 pairwin_scoring: Union[str, Callable] = 'winning_votes',
                 ):
        self.pairwin_scoring = votelib.component.pairwin_scorer.construct(
            pairwin_scoring
        )

    def evaluate(self,
                 votes: Dict[Tuple[Candidate, Candidate], int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by the ranked pairs method.

        :param votes: Condorcet votes (counts of candidate pairs as they appear
            in the voter rankings); use
            :class:`votelib.convert.RankedToCondorcetVotes`
            to produce them from ranked votes.
        :param n_seats: Number of candidates to select.
        '''
        scored = self.pairwin_scoring(votes)
        pairwise_winners = list(votes.keys())
        pairwise_winners.sort(key=votes.get, reverse=True)
        pairwise_winners.sort(key=scored.get, reverse=True)
        locked_pairs = self._lock_pairs(pairwise_winners)
        ranking = self._build_ranking(locked_pairs)
        return ranking[:n_seats]

    @classmethod
    def _lock_pairs(cls,
                    pairs: List[Tuple[Candidate, Candidate]]
                    ) -> List[Candidate]:
        locked_pairs = []
        for pair in pairs:
            if not cls._is_path(locked_pairs, pair[1], pair[0]):
                locked_pairs.append(pair)
        return locked_pairs

    @staticmethod
    def _is_path(pairs: List[Tuple[Candidate, Candidate]],
                 source: Candidate,
                 sink: Candidate,
                 ) -> bool:
        visited = set([source])
        while True:
            last_len = len(visited)
            for from_cand, to_cand in pairs:
                if from_cand in visited and to_cand not in visited:
                    visited.add(to_cand)
                    if to_cand == sink:
                        return True
            if len(visited) == last_len:
                return False

    @staticmethod
    def _build_ranking(locked_pairs: List[Tuple[Candidate, Candidate]]
                       ) -> List[Candidate]:
        edges = locked_pairs[:]
        ranking = []
        while edges:
            winners = set()
            losers = set()
            for winner, loser in edges:
                winners.add(winner)
                losers.add(loser)
            winners.difference_update(losers)
            if len(winners) != 1:
                raise votelib.evaluate.core.VotingSystemError
            winner = winners.pop()
            ranking.append(winner)
            edges = [edge for edge in edges if edge[0] != winner]
        return ranking + [next(
            node for pair in locked_pairs for node in pair
            if node not in ranking
        )]


EVALUATORS = {
    'rankedpairs_winvotes': RankedPairs(),
    'rankedpairs_margins': RankedPairs('margins'),
    'rankedpairs_pwo': RankedPairs('pairwise_opposition'),
    'copeland_2o': Copeland(),
    'copeland_raw': Copeland(second_order=False),
    'schulze': Schulze(),
    'kemeny_young': KemenyYoung(),
    'minimax_winvotes': MinimaxCondorcet(),
    'minimax_margins': MinimaxCondorcet('margins'),
    'minimax_pwo': MinimaxCondorcet('pairwise_opposition'),
}



'''Electoral threshold evaluators and other seatless selectors.

These evaluators mainly serve as auxiliary components (usually preconditions)
in proportional or similar voting systems. They are seatless selectors, which
means they return a list of elected candidates without being given the number
of them to select; that number is determined by other means.
'''

from fractions import Fraction
from typing import Any, List, Dict, Optional
from numbers import Number


import votelib.util
import votelib.evaluate.core
from votelib.candidate import Candidate, ElectionParty
from votelib.persist import simple_serialization


@simple_serialization
class AbsoluteThreshold:
    '''Absolute threshold seatless selector.

    Selects all candidates with more (or equally many) votes than the specified
    absolute number. Does not accept a number of seats as an argument to the
    ``evaluate()`` method.

    :param threshold: The absolute threshold as a number of votes.
    :param accept_equal: Whether to elect candidates that only just reach the
        threshold.
    '''
    def __init__(self,
                 threshold: Number,
                 accept_equal: bool = True,
                 ):
        self.threshold = threshold
        self.accept_equal = accept_equal

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 ) -> List[Candidate]:
        '''Select candidates by a given absolute threshold of votes.

        :param votes: Simple votes.
        '''
        return [
            cand for cand, n_votes in votelib.util.sorted_votes(votes)
            if (
                n_votes > self.threshold
                or self.accept_equal and n_votes == self.threshold
            )
        ]


@simple_serialization
class RelativeThreshold:
    '''Relative threshold seatless selector.

    Selects all candidates with more (or equally many) votes than the specified
    fraction of total votes. Does not accept a number of seats as an argument
    to the ``evaluate()`` method.

    This is a common component in proportional systems that excludes very small
    parties to increase stability of the resulting elected body.

    :param threshold: The relative threshold as a fraction of total votes.
    :param accept_equal: Whether to elect candidates that only just reach the
        threshold.
    '''
    def __init__(self,
                 threshold: Number,
                 accept_equal: bool = True,
                 ):
        self.threshold = threshold
        self.accept_equal = accept_equal

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 ) -> List[Candidate]:
        '''Select candidates by a given threshold of fraction of total votes.

        :param votes: Simple votes.
        '''
        total = sum(votes.values())
        return [
            cand for cand, n_votes in votelib.util.sorted_votes(votes)
            if (
                Fraction(n_votes, total) > self.threshold
                or self.accept_equal and n_votes == self.threshold
            )
        ]


@simple_serialization
class CoalitionMemberBracketer:
    '''Dispatch to different seatless selectors for coalitions.

    In many proportional systems the threshold to exclude smaller parties is
    raised for coalitions depending on the number of their members, to prevent
    the fragmentation of the resulting elected body.

    :param evaluators: A dictionary mapping the number of coalition members
        to the corresponding seatless selector such as
        :class:`RelativeThreshold`. Atomic parties (non-coalitions) will be
        dispatched to 1.
    :param default: A default seatless selector for coalitions that do not have
        the corresponding count in evaluators. This is useful for clauses like
        "four and more party coalitions".
    '''
    def __init__(self,
                 evaluators: Dict[int, votelib.evaluate.core.SeatlessSelector],
                 default: votelib.evaluate.core.SeatlessSelector,
                 ):
        self.evaluators = evaluators
        self.default = default

    def evaluate(self,
                 votes: Dict[ElectionParty, Number],
                 ) -> List[ElectionParty]:
        '''Select parties by dispatching to partial selectors.

        :param votes: Simple votes for parties. The keys in the dictionary must
            provide an ``is_coalition`` property and if its value is truthy, a
            ``get_n_coalition_members()`` method that returns the number of
            coalition members as an integer.
        '''
        n_member_dict = {
            cand: cand.get_n_coalition_members() if cand.is_coalition else 1
            for cand, _ in votelib.util.sorted_votes(votes)
        }
        n_member_variants = frozenset(n_member_dict.values())
        passed = {
            n_members: self.evaluators.get(
                n_members, self.default
            ).evaluate(votes)
            for n_members in n_member_variants
        }
        return [
            cand for cand, n_members in n_member_dict.items()
            if cand in passed[n_members]
        ]


@simple_serialization
class PropertyBracketer:
    '''Dispatch to different seatless selectors for some types of parties.

    In many proportional systems the threshold to exclude smaller parties is
    lowered or nonexistent for parties with a special designation, such as
    parties of regional minorities. This wrapper class allows to provide
    special seatless selectors for such special cases.

    :param property: The property (attribute) of the candidate to get as the
        key to distinguish which evaluator to use. To get the property, if
        the candidate object has a *properties* attribute, it is used,
        otherwise, ``getattr()`` is used on the candidate object directly.
    :param evaluators: A dictionary mapping the values of the property
        to the corresponding seatless selector such as
        :class:`RelativeThreshold`.
    :param default: A default seatless selector for candidates that do not
        define the specified property or whose property value is outside the
        set of keys of the evaluators dictionary.
    '''
    def __init__(self,
                 property: str,
                 evaluators: Dict[
                     Any, Optional[votelib.evaluate.core.SeatlessSelector]
                 ],
                 default: Optional[
                     votelib.evaluate.core.SeatlessSelector
                 ] = None,
                 ):
        self.property = property
        self.evaluators = evaluators
        self.default = default

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 ) -> List[Candidate]:
        '''Select candidates by dispatching to partial selectors.

        :param votes: Simple votes. The keys in the dictionary (i.e.
            candidate objects) should provide the property name specified
            at the setup of this elector; if they do not, the default evaluator
            is used for them.
        '''
        variants = {}
        results = []
        for cand, _ in votelib.util.sorted_votes(votes):
            if hasattr(cand, 'properties'):
                cand_var = cand.properties.get(self.property, NotImplemented)
            else:
                cand_var = getattr(cand, self.property, NotImplemented)
            if cand_var not in variants:
                var_eval = self.evaluators.get(cand_var, self.default)
                if var_eval:
                    variants[cand_var] = var_eval.evaluate(votes)
                else:
                    variants[cand_var] = list(votes.keys())
            if cand in variants[cand_var]:
                results.append(cand)
        return results


@simple_serialization
class AlternativeThresholds:
    '''An OR function for threshold evaluators.

    Wraps multiple seatless selectors and selects a candidate that is selected
    by any single one of them.

    :param partials: The selectors to wrap.
    '''
    # an OR function for eliminators, primarily
    def __init__(self,
                 partials: List[votelib.evaluate.core.SeatlessSelector],
                 ):
        for partial in partials:
            if votelib.evaluate.core.accepts_seats(partial):
                raise ValueError(f'seat-based evaluator {partial} in {self}')
        self.partials = partials

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 prev_gains: Dict[Candidate, int] = {},
                 ) -> List[Candidate]:
        '''Select candidates by dispatching to partial selectors and ORing.

        :param votes: Simple votes.
        :returns: A list of candidates that were selected by any single one
            of the internal partial selectors. They are ordered by mean rank
            in those partial selections.
        '''
        partial_results = []
        for partial in self.partials:
            if votelib.evaluate.core.accepts_prev_gains(partial):
                partial_result = partial.evaluate(votes, prev_gains=prev_gains)
            else:
                partial_result = partial.evaluate(votes)
            partial_results.append(partial_result)
        all_results = frozenset(
            cand for res in partial_results for cand in res
        )

        def mean_rank(cand):
            return Fraction(
                sum(
                    res.index(cand) if cand in res else len(res)
                    for res in partial_results
                ),
                len(partial_results)
            )

        return list(sorted(all_results, key=mean_rank))


@simple_serialization
class PreviousGainThreshold:
    '''A threshold on gained seats in previous election rounds.

    In some multi-round systems (especially mixed-member proportional ones),
    an alternative to clearing a national-level vote fraction threshold is to
    gain a specific minimum number of seats in the first round (e.g. in Germany
    a party may qualify for list seats if they gain three or more district
    seats).

    This evaluator passes the ``prev_gains`` argument of ordinary distribution
    evaluators to the ``votes`` argument of a given seatless selector, allowing
    one to specify a threshold based on previously gained seats.

    :param selector: The selector to evaluate the threshold on previous gains;
        :class:`AbsoluteThreshold` would be the most typical choice.
    '''
    def __init__(self, selector: votelib.evaluate.core.SeatlessSelector):
        self.selector = selector

    def evaluate(self,
                 votes: Dict[Any, int],
                 prev_gains: Dict[Candidate, int],
                 ) -> List[Candidate]:
        '''Select candidates by previous gain according to the inner selector.

        :param votes: Will be disregarded.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds. Will be passed as votes to the inner selector.
        '''
        return self.selector.evaluate(prev_gains)



'''Common functionality for components.

Functions to build function or class registers and retrievers around them.
There should normally be no need to use these functions directly.
'''

from typing import Callable, Dict, Union


def marker(register: Dict[str, Callable],
           name: str,
           signature,
           ) -> Callable[[Callable], Callable]:
    '''A registration decorator factory.'''
    def mark_function(func):
        register[func.__name__] = func
        return func
    return mark_function


def getter(register: Dict[str, Callable],
           name: str,
           signature,
           ) -> Callable[[Callable], Callable]:
    '''A register retriever factory.'''
    def get(func_def: str) -> signature:
        f'''Return a {name} function by its name.'''
        try:
            return register[func_def]
        except KeyError:
            raise KeyError(f'unknown {name}: {func_def}')
    return get


def constructer(register: Dict[str, Callable],
                name: str,
                signature,
                ) -> Callable[[Callable], Callable]:
    '''A register implicit retriever/passthrough function factory.'''
    get = getter(register, name, signature)

    def construct(func_def: Union[str, signature]
                  ) -> signature:
        f'''Construct a {name} function.

        Get a {name} function by its name from the register. If a custom
        callable is given, pass it through unchanged.
        '''
        return func_def if hasattr(func_def, '__call__') else get(func_def)

    return construct


def register_functions(*args, **kwargs):
    '''Construct the marker, getter and constructer functions at one call.'''
    return (
        marker(*args, **kwargs),
        getter(*args, **kwargs),
        constructer(*args, **kwargs),
    )



'''Various utility functions for other modules of Votelib.

There should normally be no need to use these functions directly.
'''

import operator
import itertools
import collections
import bisect
import random
from fractions import Fraction
from typing import Any, List, Tuple, Dict, Iterable
from numbers import Number

from votelib.vote import RankedVoteType
from votelib.candidate import Candidate


def add_dict_to_dict(dict1: Dict[Any, Number],
                     dict2: Dict[Any, Number],
                     ) -> None:
    for key, addition in dict2.items():
        dict1[key] = dict1.get(key, 0) + addition


def sum_dicts(dict1: Dict[Any, Number],
              dict2: Dict[Any, Number],
              ) -> Dict[Any, Number]:
    summed = dict1.copy()
    add_dict_to_dict(summed, dict2)
    return summed


def descending_dict(d: Dict[Any, Number]) -> Dict[Any, Number]:
    return dict(sorted(d.items(), key=operator.itemgetter(1), reverse=True))


def all_ranked_candidates(votes: Dict[RankedVoteType, Any]
                          ) -> List[Candidate]:
    '''Return a list of all candidates appearing in any of the rankings.

    Preserves the input ordering (i.e. first, the candidates ranked in any
    first choice position are listed in the order of the first such vote,
    first vote are listed, then all remaining candidates from the second vote,
    etc.)

    :param votes: Ranked votes.
    :returns: All unique candidates from the ranked votes.
    '''
    output = []
    for cand, rank_i, n_votes in all_rankings(votes):
        if cand not in output:
            output.append(cand)
    return output


def all_rankings(votes: Dict[RankedVoteType, Any]
                 ) -> Iterable[Tuple[Candidate, int, Any]]:
    rank_i = 0
    while True:
        used = False
        for ranking, n_votes in votes.items():
            if len(ranking) > rank_i:
                used = True
                positioned = ranking[rank_i]
                if isinstance(positioned, collections.abc.Set):
                    for cand in positioned:
                        yield cand, rank_i, n_votes
                else:
                    yield positioned, rank_i, n_votes
        if used:
            rank_i += 1
        else:
            break


def distribution_to_selection(d: Dict[Any, Number]) -> List[Any]:
    return list([cand for cand, _ in sorted(
        d.items(),
        key=operator.itemgetter(1),
        reverse=True    # rank in descending order of votes/scores/seats
    )])


def sorted_votes(votes: Dict[Any, Number],
                 descending: bool = True,
                 ) -> List[Tuple[Any, Number]]:
    '''Return votes items sorted by value.'''
    return list(sorted(
        votes.items(),
        key=operator.itemgetter(1),
        reverse=descending
    ))


def select_n_random(votes: Dict[Any, Number],
                    n: int = 1,
                    ) -> List[Any]:
    candidates, weights = zip(*sorted_votes(votes))
    candidates = list(candidates)
    cum_weights = list(itertools.accumulate(weights))
    weight_total = cum_weights[-1]
    if isinstance(weight_total, int):    # we have all integers
        return _select_n_random_int(candidates, cum_weights, n)
    elif isinstance(weight_total, Fraction):  # we have fractions, still exact
        last_denom = weight_total.denominator
        return _select_n_random_int(
            candidates, [w * last_denom for w in cum_weights], n
        )
    else:
        # TODO raise inexact arithmetics warning
        return _select_n_random_float(candidates, cum_weights, n)


def _select_n_random_int(candidates: List[Any],
                         cum_weights: List[int],
                         n: int,
                         ) -> List[Any]:
    if n > len(candidates):
        return candidates
    chosen = []
    while len(chosen) < n:
        new_cand_i = bisect.bisect_left(
            cum_weights,
            random.randrange(1, cum_weights[-1] + 1)
        )
        chosen.append(candidates.pop(new_cand_i))
        subtract_wt = cum_weights.pop(new_cand_i)
        if new_cand_i != 0 and cum_weights:
            subtract_wt -= cum_weights[new_cand_i-1]
        cum_weights = (
            cum_weights[:new_cand_i]
            + [wt - subtract_wt for wt in cum_weights[new_cand_i:]]
        )
    return chosen


def _select_n_random_float(candidates: List[Any],
                           cum_weights: List[Number],
                           n: int,
                           ) -> List[Any]:
    return random.choices(
        candidates,
        cum_weights=cum_weights,
        k=n
    )


def exact_mean(values: List[Number]) -> Number:
    total = sum(values)
    if isinstance(total, float):
        return total / len(values)
    else:
        return Fraction(total, len(values))


EXACT_AGGREGATORS = {
    'mean': exact_mean,
}



"""Shared functionality for ballot/election file I/O. Internal."""

import typing
from typing import Any, Tuple, Callable, Iterable, TextIO, TypeVar

FilePayload = TypeVar('FilePayload')


class NotSupportedInFormat(Exception):
    """Signals that the given element is not supported by the I/O format."""

    FORMAT: str = NotImplemented

    def __init__(self, what: str):
        super().__init__(f'{what} not supported by {self.FORMAT}')


class ParseError(Exception):
    """An input that is invalid according to the given format was detected."""
    pass


def loaders(line_loader: Callable[..., FilePayload]
            ) -> Tuple[Callable[..., FilePayload], Callable[..., FilePayload]]:
    """Create load() and loads() functions from an iterating function."""
    return_annot = typing.get_type_hints(line_loader).get('return')
    if return_annot is None:
        return_annot = Any

    def load(file: TextIO, **kwargs) -> return_annot:
        return line_loader(file, **kwargs)

    def loads(text: str, **kwargs) -> return_annot:
        return line_loader(iter(text.split('\n')), **kwargs)

    return load, loads


def dumpers(line_dumper: Callable[..., Iterable[str]]
            ) -> Tuple[Callable[..., None], Callable[..., str]]:
    """Create dump() and dumps() functions from a line generator function."""

    def dump(blt_file: TextIO, *args, **kwargs) -> None:
        for line in line_dumper(*args, **kwargs):
            if not line.endswith('\n'):
                line += '\n'
            blt_file.write(line)

    def dumps(*args, **kwargs) -> str:
        return ''.join(
            line + ('' if line.endswith('\n') else '\n')
            for line in line_dumper(*args, **kwargs)
        )

    return dump, dumps



'''Votelib - a library for evaluating election results.

Votelib objects provide the means to evaluate elections under many known
election systems, be it simple or complicated, obscure or ubiquitous.

An election system usually specifies the following:

-   Who can stand for the election (what is a valid nomination). This can be
    checked using the nominators from the ``candidate`` module.
-   What forms of votes are valid. This can be checked using the vote
    validators from the ``vote`` module for different vote types from ranked
    to range votes.
-   How to determine who is elected. This is the task of the ``evaluate``
    subpackage and its many modules, which contains evaluators for most of the
    world's known systems. Many of those considered standalone evaluation
    methods are actually reducible variants that can use an another evaluator;
    for these cases, the ``convert`` module provides converters for votes and
    results to create more complex systems.

The evaluator can be combined with the validator and nominator by the machinery
in the :mod:`evaluate` subpackage. The :class:`VotingSystem` object can then
wrap it into a formalized and named election system.
'''

import votelib.evaluate
from votelib.persist import simple_serialization    # noqa: F401


@simple_serialization
class VotingSystem:
    '''A named voting system. Wraps an election evaluator.

    :param name: Name of the system; usually mainly includes the body or
        position to be elected.
    :param evaluator: Evaluator representing the system. Can be combined with
        the validator and nominator by machinery in the :mod:`evaluate`
        subpackage.
    '''
    def __init__(self, name: str, evaluator: votelib.evaluate.Evaluator):
        self.name = name
        self.evaluator = evaluator

    def evaluate(self, *args, **kwargs):
        '''Return the evaluator's results of the system for the votes given.'''
        return self.evaluator.evaluate(*args, **kwargs)



'''Vote type specifications and vote validators.

Vote types vary from system to system and are only loosely tied to the method
of evaluation (usually, the richer vote types can be reduced to use simple
evaluators but not vice versa). The following vote types are recognized by
Votelib:

-   **Simple** votes - a voter votes for a single candidate. Represented by the
    candidate object itself.
-   **Approval** votes - a voter selects a number of candidates and votes for
    them equally. Represented by a frozen set of candidate objects.
-   **Ranked** votes - a voter ranks a number of candidates. Represented by a
    tuple of candidate objects or frozen sets of them (to account for possible
    tied rankings).
-   **Score** votes - a voter assigns a score (label) to a number of
    candidates. The scores might either be selected from a predefined set or
    arbitrary numbers from a range.

Voting systems generally require one of the above types and usually impose
additional restrictions such as minimum and maximum number of candidates voted
for or ranked. These restrictions are handled by parameters accepted by the
vote validators.

Vote validators validate individual votes, i.e. the keys of the dictionaries
that should be passed to the evaluators. If a vote is invalid, they raise a
subclass of :class:`VoteError` (or :class:`CandidateError`, if a candidate
contained in the vote is invalid).
Some converter objects wrap these validators and catch these errors to remove
invalid votes, for example.
'''

import abc
from typing import Any, Tuple, FrozenSet, Dict, Union, Optional, Collection
from numbers import Number

import collections

import votelib.candidate
from votelib.candidate import Candidate
from votelib.persist import simple_serialization


class VoteError(Exception, metaclass=abc.ABCMeta):
    '''A vote is invalid given the election rules.'''
    pass


class VoteTypeError(VoteError):
    '''A vote is of an invalid type.

    E.g. ranked votes in place of simple votes.

    :param vtype: Vote type detected as invalid.
    :param expected: Vote type that was expected.
    '''
    def __init__(self, vtype: type, expected: type = None):
        self.vtype = vtype
        self.expected = expected
        message = f'invalid vote type: {vtype}'
        if expected:
            message += f', must be {expected}'
        super().__init__(message)


class VoteMagnitudeError(VoteError):
    '''A vote is too small or too large.

    :param value: Size of the vote that was found to be invalid.
    :param min_value: Minimum value permissible in the context.
    :param max_value: Maximum value permissible in the context.
    :param value_name: Role of the vote size (e.g. number of approval votes,
        number of ranked candidates...)
    '''
    def __init__(self,
                 value: Number,
                 min_value: Optional[Number] = None,
                 max_value: Optional[Number] = None,
                 value_name: str = 'count',
                 ):
        self.value = value
        self.min_value = min_value
        self.max_value = max_value
        message = f'invalid vote {value_name}: {value}'
        if min_value is not None or max_value is not None:
            message += ', must be '
            parts = []
            if min_value is not None:
                message += f'>={min_value}'
            if max_value is not None:
                message += f'<={max_value}'
            message += ', '.join(parts)
        super().__init__(message)


class VoteValueError(VoteError):
    '''An explicitly given vote value is invalid.

    :param value: Value of the vote that is invalid.
    :param candidate: A candidate that the vote was given for. If None, a
        specific candidate could not be pinpointed.
    :param allowed: A spectrum of values that is allowed at the given point.
    '''
    def __init__(self,
                 value: Any,
                 candidate: Optional[Candidate] = None,
                 allowed: Any = None,
                 ):
        self.value = value
        self.candidate = candidate
        self.allowed = allowed
        message = f'invalid vote: {value}'
        if candidate is not None:
            message += f' for candidate {candidate}'
        if allowed is not None:
            message += f', allowed: {allowed}'
        super().__init__(message)


SimpleVoteType = Candidate
ApprovalVoteType = FrozenSet[Candidate]
RankedVoteType = Tuple[Union[Candidate, FrozenSet[Candidate]], ...]
ScoreVoteType = FrozenSet[Tuple[Candidate, Any]]

IntBoundsTupleType = Tuple[Optional[int], Optional[int]]
NumBoundsTupleType = Tuple[Optional[Number], Optional[Number]]


class VoteMagnitudeChecker:
    '''A helper class to check if a value is in a specified range.

    :param bounds: A tuple with lower and upper bounds (inclusive) for the
        value to be checked. None means the respective bound is not checked.
    :param value_name: Name of the value to be checked (included in the error
        message).
    '''
    def __init__(self,
                 bounds: NumBoundsTupleType = (None, None),
                 value_name: str = 'count',
                 ):
        self.min_value, self.max_value = bounds
        self.value_name = value_name
        self._active = self.min_value is not None or self.max_value is not None

    def to_dict(self) -> Dict[str, Any]:
        return {
            'bounds': [self.min_value, self.max_value],
            'value_name': self.value_name
        }

    def __bool__(self) -> bool:
        '''Return True if the checker contains any constraints to check.'''
        return self._active

    def is_valid(self, value: Number) -> bool:
        '''Return True if the value is within the given range.'''
        return (
            (self.min_value is None or value >= self.min_value)
            and (self.max_value is None or value <= self.max_value)
        )

    def check(self, value: Number) -> None:
        '''Check if the value is within the given range.

        :raises VoteMagnitudeError: If the value is outside the given
            range.
        '''
        if not self.is_valid(value):
            raise VoteMagnitudeError(
                value, self.min_value, self.max_value, self.value_name
            )


class VoteValidator(metaclass=abc.ABCMeta):
    '''Validate that a single vote is valid under the election rules.

    Base class, not intended for direct use.
    '''
    @abc.abstractmethod
    def validate(self, vote: Any) -> None:
        '''Check if the vote satisfies criteria given by the voting system.

        :raises NotImplementedError:
        '''
        raise NotImplementedError


DEFAULT_NOMINATOR = votelib.candidate.BasicNominator()


@simple_serialization
class SimpleVoteValidator:
    '''Validate a simple vote (voting directly for a single candidate).

    The candidate must be a valid atomic candidate object according to the
    nominator object specified.

    :param nominator: Nominator used to check candidates. The default uses only
        technical criteria specified by the :class:`Candidate` class.
    '''
    def __init__(self,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        self.nominator = nominator

    def validate(self, vote: Candidate) -> None:
        '''Check if the candidate is valid.

        :param vote: Simple vote to be checked.
        :raises CandidateError: If the candidate is invalid.
        '''
        self.nominator.validate(vote)


@simple_serialization
class ApprovalVoteValidator:
    '''Validate an approval vote (voting for a number of candidates equally).

    The vote must be a frozen set containing valid atomic candidates. The
    atomic candidates must not be tuples.

    :param vote_count_bounds: A tuple with lower and upper bounds
        (inclusive) for the number of candidates any vote can contain.
        None means the respective bound is not checked.
        Ignored if count_checker is given.
    :param count_checker: A :class:`VoteMagnitudeChecker` that checks the
        number of candidates any vote can contain.
    :param nominator: Nominator used to check candidates. The default uses the
        technical criteria specified by the :class:`Candidate` class.
    '''
    serialize_params = ['count_checker', 'nominator']

    def __init__(self,
                 vote_count_bounds: IntBoundsTupleType = (None, None),
                 count_checker: Optional[VoteMagnitudeChecker] = None,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        if count_checker is None:
            count_checker = VoteMagnitudeChecker(vote_count_bounds)
        self.count_checker = count_checker
        self.nominator = nominator

    def validate(self, vote: FrozenSet[Candidate]) -> None:
        '''Check if the approval vote is valid.

        :param vote: Approval vote to be checked.
        :raises VoteTypeError: If the vote is not a frozen set.
        :raises CandidateError: If any of the contained candidates
            is invalid.
        :raises VoteMagnitudeError: If the number of candidates voted
            for is out of allowed bounds.
        '''
        if not isinstance(vote, frozenset):
            raise VoteTypeError(vote, frozenset)
        for item in vote:
            self.nominator.validate(item)
        self.count_checker.check(len(vote))


@simple_serialization
class RankedVoteValidator:
    '''Validate a ranked vote (ranking of a number of candidates).

    The vote must be a tuple of candidates or frozen sets thereof. Usage of
    sets indicates tied rankings, which are not common but allowed in some
    systems.

    :param total_vote_count_bounds: A tuple with lower and upper bounds
        (inclusive) for the number of candidates any vote can rank.
        None means the respective bound is not checked.
        Ignored if total_count_checker is given.
    :param rank_vote_count_bounds: A tuple with lower and upper bounds
        (inclusive) for the number of candidates allowed to share any rank.
        The default settings disallow tied rankings. None means the respective
        bound is not checked. Alternatively, a dictionary can be specified
        that maps ranks (1-indexed) to bound tuples.
        Ignored if rank_vote_count_checkers is given.
    :param total_count_checker: A :class:`VoteMagnitudeChecker` that checks the
        total number of candidates any vote can rank.
    :param rank_vote_count_checkers: A mapping of integer ranks to instances
        of :class:`VoteMagnitudeChecker` to check numbers of candidates
        allowed to share any rank.
    :param nominator: Nominator used to check candidates. The default uses only
        technical criteria specified by the :class:`Candidate` class.
    '''
    serialize_params = [
        'total_count_checker',
        'rank_vote_count_checkers',
        'nominator'
    ]

    def __init__(self,
                 total_vote_count_bounds: IntBoundsTupleType = (None, None),
                 rank_vote_count_bounds: Union[
                     IntBoundsTupleType,
                     Dict[int, IntBoundsTupleType],
                 ] = (1, 1),
                 total_count_checker: Optional[VoteMagnitudeChecker] = None,
                 rank_vote_count_checkers: Optional[
                     Dict[int, VoteMagnitudeChecker]
                 ] = None,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        if total_count_checker is None:
            total_count_checker = VoteMagnitudeChecker(total_vote_count_bounds)
        self.total_count_checker = total_count_checker
        if rank_vote_count_checkers is None:
            if hasattr(rank_vote_count_bounds, 'items'):
                rank_vote_count_checkers = collections.defaultdict(
                    lambda: VoteMagnitudeChecker((None, None))
                )
                for rank, bounds in rank_vote_count_bounds.items():
                    rank_vote_count_checkers[rank] = VoteMagnitudeChecker(
                        bounds
                    )
            else:
                rank_checker = VoteMagnitudeChecker(rank_vote_count_bounds)
                rank_vote_count_checkers = collections.defaultdict(
                    lambda: rank_checker
                )
        self.rank_vote_count_checkers = rank_vote_count_checkers
        self.nominator = nominator

    def validate(self, vote: RankedVoteType) -> None:
        '''Check if the ranked vote is valid.

        :param vote: Ranked vote to be checked.
        :raises VoteTypeError: If the vote is not a tuple.
        :raises VoteError: If any candidate is specified more than once
            in the ranking.
        :raises CandidateError: If any of the contained candidates
            is invalid.
        :raises VoteMagnitudeError: If the number of candidates (total
            or at particular ranking tier) is out of the specified bounds.
        '''
        if not isinstance(vote, tuple):
            raise VoteTypeError(vote, tuple)
        total_votes = 0
        all_candidates = set()
        for rank_i, item in enumerate(vote):
            if isinstance(item, collections.abc.Set):
                self.rank_vote_count_checkers[rank_i+1].check(len(item))
                all_candidates.update(item)
                total_votes += len(item)
            else:
                all_candidates.add(item)
                total_votes += 1
        self.total_count_checker.check(total_votes)
        if len(all_candidates) < total_votes:
            raise VoteError(f'duplicated candidates: {vote}')
        for cand in all_candidates:
            self.nominator.validate(cand)


class ScoreVoteValidator:
    # parent class for EnumScoreVoteValidator and RangeVoteValidator
    def __init__(self,
                 allowed_scorings: IntBoundsTupleType = (None, None),
                 sum_bounds: Union[
                     NumBoundsTupleType, Dict[int, NumBoundsTupleType]
                 ] = (None, None),
                 n_scorings_checker: Optional[VoteMagnitudeChecker] = None,
                 sum_checkers: Optional[
                     Dict[int, VoteMagnitudeChecker]
                 ] = None,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        if n_scorings_checker is None:
            n_scorings_checker = VoteMagnitudeChecker(allowed_scorings)
        if sum_checkers is None:
            if hasattr(sum_bounds, 'items'):
                no_sc = VoteMagnitudeChecker((None, None), 'sum')
                sum_checkers = collections.defaultdict(lambda: no_sc)
                for n_scorings, bounds in sum_bounds.items():
                    sum_checkers[n_scorings] = VoteMagnitudeChecker(
                        bounds, 'sum'
                    )
            else:
                default_sc = VoteMagnitudeChecker(sum_bounds, 'sum')
                sum_checkers = collections.defaultdict(lambda: default_sc)
        self.n_scorings_checker = n_scorings_checker
        self.sum_checkers = sum_checkers
        self.nominator = nominator

    def validate(self, vote: ScoreVoteType) -> bool:
        if not isinstance(vote, frozenset):
            raise VoteTypeError(vote, frozenset)
        n_scorings = len(vote)
        self.n_scorings_checker.check(n_scorings)
        for item in vote:
            if not isinstance(item, tuple):
                raise VoteTypeError(item, tuple)
            if not len(item) == 2:
                raise VoteMagnitudeError(
                    len(item), 2, 2, 'scoring pair length'
                )
            self.nominator.validate(item[0])
        sum_checker = self.sum_checkers[n_scorings]
        if sum_checker:
            sum_checker.check(sum(scoring[1] for scoring in vote))


@simple_serialization
class EnumScoreVoteValidator(ScoreVoteValidator):
    '''Validate an enumeration-based score vote.

    An enumeration-based score vote assigns scores from a predefined finite set
    to candidates. It should be represented as a frozen set of doubles
    containing a candidate and the associated score. This variant is the most
    common for score-based voting systems since the voters are usually given
    a predefined finite set of (possibly non-numeric) scores; for true range
    voting where the voters might specify arbitrary score values, use
    :class:`RangeVoteValidator`.

    If numeric scores are used, the usage of exact numeric types (integers,
    fractions, decimals) is encouraged.

    :param score_levels: A collection of allowed scores.
    :param allowed_scorings: A tuple with lower and upper bounds
        (inclusive) for the number of candidates allowed to appear in any
        single vote. None means the respective bound is not checked.
        Ignored if n_scorings_checker is given.
    :param sum_bounds: A tuple with lower and upper bounds
        (inclusive) for the total sum of numeric scores allowed for any
        single vote. None means the respective bound is not checked. You can
        also provide a dictionary that gives different sum bounds for different
        numbers of candidates scored (numbers of candidates scored that do not
        have a corresponding key will not have their sums checked then).
        Ignored if sum_checkers are given.
    :param n_scorings_checker: A :class:`VoteMagnitudeChecker` that checks the
        total number of candidates any vote can score.
    :param sum_checkers: A mapping of integer numbers of scored
        candidates to instances of :class:`VoteMagnitudeChecker` checking the
        allowed total score allocated to all candidates.
    :param nominator: Nominator used to check candidates. The default uses only
        technical criteria specified by the :class:`Candidate` class.
    '''
    serialize_params = [
        'score_levels',
        'n_scorings_checker',
        'sum_checkers',
        'nominator'
    ]

    def __init__(self,
                 score_levels: Collection[Any],
                 allowed_scorings: IntBoundsTupleType = (None, None),
                 sum_bounds: Union[
                     NumBoundsTupleType, Dict[int, NumBoundsTupleType]
                 ] = (None, None),
                 n_scorings_checker: Optional[VoteMagnitudeChecker] = None,
                 sum_checkers: Optional[
                     Dict[int, VoteMagnitudeChecker]
                 ] = None,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        super().__init__(
            allowed_scorings, sum_bounds,
            n_scorings_checker, sum_checkers,
            nominator
        )
        self.score_levels = list(score_levels)

    def validate(self, vote: ScoreVoteType) -> bool:
        '''Check if the enumeration-based score vote is valid.

        :param vote: Score vote to be checked.
        :raises VoteTypeError: If the vote is not a frozen set.
        :raises CandidateError: If any of the contained candidates
            is invalid.
        :raises VoteMagnitudeError: If the number of candidates scored
            or the total sum of scores is out of the specified bounds, or any
            of the items in the vote is not a tuple pair.
        :raises VoteValueError: If any of the scores is not in the
            predefined set of allowed scores.
        '''
        super().validate(vote)
        for cand, score in vote:
            if score not in self.score_levels:
                raise VoteValueError(score, cand, self.score_levels)


@simple_serialization
class RangeVoteValidator(ScoreVoteValidator):
    '''Validate a range (non-enumerative score) vote.

    An range vote assigns scores from a predefined interval to candidates.
    It should be represented as a frozen set of doubles
    containing a candidate and the associated score. This variant is mostly
    theoretical since most systems only allow choice from a finite set of
    scores; for that variant, use :class:`EnumScoreVoteValidator`.

    The usage of exact numeric types (integers, fractions, decimals) for scores
    is encouraged.

    :param range: A tuple with lower and upper bounds
        (inclusive) for any single score value. None means the respective bound
        is not checked.
    :param allowed_scorings: A tuple with lower and upper bounds
        (inclusive) for the number of candidates allowed to appear in any
        single vote. None means the respective bound is not checked.
        Ignored if n_scorings_checker is given.
    :param sum_bounds: A tuple with lower and upper bounds
        (inclusive) for the total sum of numeric scores allowed for any
        single vote. None means the respective bound is not checked.
        Ignored if sum_checkers are given.
    :param n_scorings_checker: A :class:`VoteMagnitudeChecker` that checks the
        total number of candidates any vote can score.
    :param sum_checkers: A mapping of integer numbers of scored
        candidates to instances of :class:`VoteMagnitudeChecker` checking the
        allowed total score allocated to all candidates.
    :param nominator: Nominator used to check candidates. The default uses only
        technical criteria specified by the :class:`Candidate` class.
    '''
    serialize_params = [
        'range_checker',
        'n_scorings_checker',
        'sum_checkers',
        'nominator'
    ]

    def __init__(self,
                 range: NumBoundsTupleType = (None, None),
                 allowed_scorings: IntBoundsTupleType = (None, None),
                 sum_bounds: Union[
                     NumBoundsTupleType, Dict[int, NumBoundsTupleType]
                 ] = (None, None),
                 range_checker: Optional[VoteMagnitudeChecker] = None,
                 n_scorings_checker: Optional[VoteMagnitudeChecker] = None,
                 sum_checkers: Optional[
                     Dict[int, VoteMagnitudeChecker]
                 ] = None,
                 nominator: votelib.candidate.Nominator = DEFAULT_NOMINATOR,
                 ):
        super().__init__(
            allowed_scorings, sum_bounds,
            n_scorings_checker, sum_checkers,
            nominator
        )
        if range_checker is None:
            range_checker = VoteMagnitudeChecker(range, 'range vote value')
        self.range_checker = range_checker

    def validate(self, vote: ScoreVoteType) -> bool:
        '''Check if the range vote is valid.

        :param vote: Range (score) vote to be checked.
        :raises VoteTypeError: If the vote is not a frozen set.
        :raises CandidateError: If any of the contained candidates
            is invalid.
        :raises VoteMagnitudeError: If the number of candidates scored,
            the total sum of scores or any single score is out of the specified
            bounds, or any of the items in the vote is not a tuple pair.
        '''
        super().validate(vote)
        for cand, score in vote:
            self.range_checker.check(score)


class VoteSubsetter(metaclass=abc.ABCMeta):
    '''Abstract class for vote subsetters.

    Vote subsetters reduce a single vote (object) to a smaller variant in which
    only a specified subset of candidates is featured.
    This is trivial for simple votes but is important to do properly for ranked
    votes.

    Vote subsetters should return None if the vote is to be discarded in the
    subset.

    Subsetted votes are handy in some cases such as tiebreaking. If you need to
    subset an entire voting dictionary, wrap the vote subsetter into an
    :class:`votelib.convert.SubsettedVotes` instance.
    '''
    @abc.abstractmethod
    def subset(self,
               vote: Any,
               subset: Collection[Candidate],
               ) -> Union[Any, None]:
        raise NotImplementedError


@simple_serialization
class SimpleSubsetter:
    '''A subsetter for simple votes.'''

    def subset(self,
               vote: Candidate,
               subset: Collection[Candidate],
               ) -> Union[Candidate, None]:
        '''Return vote if it is in subset, else None.'''
        return vote if vote in subset else None


@simple_serialization
class ApprovalSubsetter:
    '''A subsetter for approval votes.'''

    def subset(self,
               vote: FrozenSet[Candidate],
               subset: Collection[Candidate],
               ) -> FrozenSet[Candidate]:
        '''Return a subset of the approval vote by intersection.

        :returns: An intersection of the vote with the candidate subset.
        '''
        return vote.intersection(subset)


@simple_serialization
class RankedSubsetter:
    '''A subsetter for ranked votes.'''
    # TODO: implement keeping ranks as skipped

    def subset(self,
               vote: RankedVoteType,
               subset: Collection[Candidate],
               ) -> RankedVoteType:
        '''Return a ranked vote ranking only the candidates in the subset.

        Ranks that contain only candidates outside the subset are removed and
        the list is shortened. Shared ranks that only have one of their
        candidates in the subset are converted to simple ranks.

        :returns: A ranked vote ranking only the candidates in the subset.
        '''
        sub_ranking = []
        for rank in vote:
            if isinstance(rank, collections.abc.Set):
                sub_rank = rank.intersection(subset)
                if sub_rank:
                    if len(sub_rank) == 1:
                        sub_ranking.append(next(iter(sub_rank)))
                    else:
                        sub_ranking.append(sub_rank)
            elif rank in subset:
                sub_ranking.append(rank)
        return tuple(sub_ranking)


@simple_serialization
class ScoreSubsetter:
    '''A subsetter for score votes.'''

    def subset(self,
               vote: ScoreVoteType,
               subset: Collection[Candidate],
               ) -> ScoreVoteType:
        '''Return a score vote scoring only the candidates in the subset.

        Scores for candidates outside the subset are removed.

        :returns: A score vote scoring only the candidates in the subset.
        '''
        sub_score = set()
        for cand, score in vote:
            if cand in subset:
                sub_score.add((cand, score))
        return frozenset(sub_score)



"""Read and write STV format files.

The STV (or STV.pm) files contain cast votes in a ranked election and,
optionally, an exact definition of an election system. [#stvlob]_

.. [#wpav] "STV File Format".
    https://lobitos.net/voting/format.html
"""
import re
import math
import operator
import fractions
import decimal
import collections
import warnings
from typing import List, Tuple, Dict, Union, Iterable, Optional, Collection
from numbers import Number

import votelib
import votelib.convert
import votelib.evaluate
import votelib.evaluate.auxiliary
import votelib.evaluate.sequential
import votelib.io.blt
import votelib.io.core
import votelib.util
import votelib.component.transfer as transfer
from votelib.candidate import Candidate
from votelib.evaluate.sequential import \
    TransferableVoteSelector, TransferableVoteDistributor


SUPPORTED_QUOTAS: List[str] = ['droop', 'hare']


class NotSupportedInSTV(votelib.io.core.NotSupportedInFormat):
    FORMAT = 'STV file'


class STVParseError(votelib.io.core.ParseError):
    pass


def dump_lines(votes: Dict[Tuple[Candidate, ...], Number],
               system: Union[
                   votelib.VotingSystem,
                   votelib.evaluate.Evaluator,
                   None
               ] = None,
               candidates: Optional[List[Candidate]] = None,
               n_seats: Optional[int] = None,
               output_method: bool = True,
               ) -> Iterable[str]:
    """Dump the election data into an STV.pm file.

    :param votes: Ranked votes for the election.
    :param system: The election system (or just an evaluator) used in the
        election.
    :param candidates: List of candidates participating in the election.
        If None, will be inferred from votes.
    :param n_seats: Number of seats contested. Supplying this along with a
        fixed seat count evaluator will result in a duplicate marking in the
        output.
    :param output_method: If True, output the detailed parameters of the
        election system used. Only applies when system is given.
        Setting this to False will omit the vote transfer setup, but it allows
        to output partial STV setups from election systems that are not
        supported by STV files and would thus normally error out (such as
        non-uniform elimination). This parameter does not affect quota setup
        output, which is always attempted.
    """
    if system is not None:
        yield from _dump_system(system, output_method=output_method)
        if n_seats is not None:
            yield f'seats={n_seats}'
        yield from _dump_ballots(votes, candidates)
    else:
        yield 'method=blt'
        yield 'ballots=blt'
        yield from votelib.io.blt.dump_lines(
            votes, n_seats, candidates=candidates
        )


dump, dumps = votelib.io.core.dumpers(dump_lines)


def _dump_system(system: Union[
                     votelib.VotingSystem,
                     votelib.evaluate.Evaluator
                 ],
                 **kwargs) -> Iterable[str]:
    if isinstance(system, votelib.VotingSystem):
        yield f'title={system.name}'
        yield from _dump_system(system.evaluator, **kwargs)
    elif isinstance(system, votelib.evaluate.FixedSeatCount):
        yield f'seats={system.n_seats}'
        yield from _dump_system(system.evaluator, **kwargs)
    elif isinstance(system, votelib.evaluate.TieBreaking):
        yield from _dump_system(system.main, **kwargs)
        yield from _dump_tiebreaker(system.tiebreaker)
    elif isinstance(system, TransferableVoteDistributor):
        warnings.warn('TransferableVoteDistributor class will not be preserved'
                      ' in STV file, will store TransferableVoteSelector'
                      ' instead')
        yield from _dump_tveval(system, **kwargs)
    elif isinstance(system, TransferableVoteSelector):
        yield from _dump_tveval(system._inner, **kwargs)


def _dump_tveval(evaluator: TransferableVoteDistributor,
                 output_method: bool = True,
                 ) -> Iterable[str]:
    if output_method:
        if evaluator.retainer is not None:
            raise NotSupportedInSTV(f'transferable vote retainer'
                                    f'{evaluator.retainer}')
        if evaluator.eliminate_step != -1:
            raise NotSupportedInSTV(f'elimination {evaluator.eliminate_step}')
        if isinstance(evaluator.transferer, transfer.Gregory):
            yield 'method=BC'
        else:
            raise NotSupportedInSTV(f'transfer method {evaluator.transferer}')
    if hasattr(evaluator.quota_function, '__name__'):
        quota_name = evaluator.quota_function.__name__
        if quota_name in SUPPORTED_QUOTAS:
            yield f'quota={evaluator.quota_function.__name__}'
        else:
            raise NotSupportedInSTV(f'quota {quota_name}')
    if getattr(evaluator, 'mandatory_quota', False):
        yield 'quota=mandatory'


ORDER_SELECTORS = (
    votelib.evaluate.auxiliary.InputOrderSelector,
    votelib.evaluate.auxiliary.CandidateNumberRanker,
)


def _dump_tiebreaker(evaluator: votelib.evaluate.Evaluator) -> Iterable[str]:
    if isinstance(evaluator, votelib.evaluate.PreConverted):
        if type(evaluator.converter) not in votelib.convert.RANKED_TO_SIMPLE:
            raise NotSupportedInSTV(f'conversion {evaluator.converter}')
        yield from _dump_tiebreaker(evaluator.evaluator)
    elif isinstance(evaluator, ORDER_SELECTORS):
        yield 'random=non'
    elif isinstance(evaluator, votelib.evaluate.auxiliary.Sortitor):
        if evaluator.seed is not None:
            yield f'random={evaluator.seed}'
    else:
        raise NotSupportedInSTV(f'tiebreaker {evaluator}')


def _dump_ballots(votes: Dict[Tuple[Candidate, ...], Number],
                  candidates: Optional[List[Candidate]] = None,
                  ) -> Iterable[str]:
    # unordered format supported only
    if candidates is None:
        candidates = votelib.util.all_ranked_candidates(votes)
    cand_names = {cand: _candidate_name(cand) for cand in candidates}
    cand_nicks = _candidate_nicks(cand_names)
    for cand in candidates:
        prefix = 'candidate'
        if getattr(cand, 'withdrawn', False):
            prefix = 'withdrawn'
        yield f'{prefix}={cand_nicks[cand]} {cand_names[cand]}'
    yield f'ballots={len(votes)}'
    for ranking, n_votes in votes.items():
        multiplier = f'{n_votes}X ' if n_votes != 1 else ''
        yield multiplier + _ranking_to_str(ranking, cand_nicks)
    yield 'end'


def _ranking_to_str(ranking: Tuple[Candidate, ...],
                    nicks: Dict[Candidate, str],
                    ) -> str:
    try:
        return ' '.join(nicks[cand] for cand in ranking)
    except KeyError as err:
        if err.args and isinstance(err.args[0], collections.abc.Set):
            raise NotSupportedInSTV(f'equal rankings: {ranking}')
        else:
            raise err


def _candidate_name(candidate: Candidate) -> str:
    if isinstance(candidate, str):
        return candidate
    elif getattr(candidate, 'name', None) is not None:
        return candidate.name
    else:
        return str(candidate)


def _candidate_nicks(cand_names: Dict[Candidate, str]) -> Dict[Candidate, str]:
    all_initials = []
    for cand_name in cand_names.values():
        cand_initials = _name_to_initials(cand_name)
        if cand_initials in all_initials:    # duplicate, fall back to trivial
            return _ordinal_candidate_nicks(cand_names.keys())
        else:
            all_initials.append(cand_initials)
    return dict(zip(cand_names.keys(), all_initials))


def _name_to_initials(name: str) -> str:
    return ''.join(part[0].lower() for part in re.split(r'\W', name))


def _ordinal_candidate_nicks(cand_names: Collection[Candidate]
                             ) -> Dict[Candidate, str]:
    n_letters = int(math.ceil(math.log(len(cand_names)) / math.log(26)))
    nicks = {}
    for cand_i, cand in enumerate(cand_names):
        nick_letters = []
        for letter_i in range(n_letters):
            nick_letters.append(chr(97 + cand_i % 26))
            cand_i //= 26
        nicks[cand] = ''.join(nick_letters)
    return nicks


def load_lines(lines: Iterable[str]) -> Tuple[
                   Dict[Tuple[Candidate, ...], Number],
                   Optional[votelib.VotingSystem],
                   List[Candidate],
               ]:
    system, candidates, nicks, n_ballots, is_ordered = _load_system(lines)
    if n_ballots is None:
        # BLT mode invoked, the rest of the file is in BLT format.
        try:
            votes, blt_n_seats, blt_candidates, election_name = \
                votelib.io.blt.load_lines(lines)
        except votelib.io.blt.BLTParseError as err:
            raise STVParseError('BLT content parsing failed') from err
        # Replace the STV config results with the BLT data where applicable.
        if system.name is None and election_name:
            system.name = election_name
        system.evaluator = votelib.evaluate.FixedSeatCount(
            system.evaluator, blt_n_seats
        )
        if not candidates and blt_candidates:
            candidates = blt_candidates
    else:
        loader = _load_ordered_votes if is_ordered else _load_unordered_votes
        votes = loader(_iter_vote_lines(lines, n_ballots), nicks)
    return votes, system, candidates


def _load_system(lines: Iterable[str]) -> Tuple[
                     Optional[votelib.VotingSystem],
                     List[Candidate],
                     Dict[str, Candidate],
                     Optional[int],
                     bool,
                 ]:
    syscomps = {}
    candidates = []
    nicks = {}
    nick_orders = []
    for line in lines:
        key, value = _parse_header_line(line)
        if key is None:
            continue    # comment or empty line
        elif key == 'ballots':
            if nick_orders:
                # reorder nicks into order= spec
                nicks = {nick: nicks[nick] for nick in nick_orders}
            return (
                _create_system(**syscomps), candidates, nicks,
                _parse_n_ballots(value), bool(nick_orders)
            )
        elif key == 'order':
            nick_orders = value.split()
        elif key in ('candidate', 'withdrawn'):
            nick, name = value.split(None, 1)
            cand = votelib.candidate.Person(
                name,
                number=len(candidates)+1,
                withdrawn=(key == 'withdrawn'),
            )
            candidates.append(cand)
            nicks[nick] = cand
        elif key in syscomps:
            syscomps[key] = (syscomps[key], value)
        else:
            syscomps[key] = value
    raise STVParseError('end of file before ballot data')


def _parse_header_line(line: str) -> Tuple[str, str]:
    if '#' in line:
        line = line[:line.find('#')]
    line = line.strip()
    if not line:
        return None, None
    if '=' in line:
        return line.split('=', 1)
    else:
        raise STVParseError(f'invalid STV header line: {line!r}')


def _parse_n_ballots(value: str) -> Optional[int]:
    if value == 'blt':
        return None
    elif value.isdigit():
        return int(value)
    else:
        raise STVParseError(f'invalid ballot count: {value!r}')


def _iter_vote_lines(lines: Iterable[str],
                     n_ballots: int,
                     ) -> Iterable[Tuple[Number, List[str]]]:
    has_ended = False
    for i, line in enumerate(lines):
        line = line.strip()
        if line == 'end':
            if i != n_ballots:
                raise STVParseError(f'expected {n_ballots} ballots, got {i}')
            has_ended = True
            break
        if not line:
            continue
        items = line.split()
        if items[0].endswith('X'):
            yield _parse_multiplier(items[0][:-1], i), items[1:]
        else:
            yield 1, items
    if not has_ended:
        raise STVParseError('no "end" terminator line')


def _parse_multiplier(mult: str, line_i: int) -> Number:
    inner_err = None
    try:
        if '/' in mult:
            return fractions.Fraction(mult)
        elif '.' in mult:
            return decimal.Decimal(mult)
        elif mult.isdigit():
            return int(mult)
    except ValueError as err:
        inner_err = err
    parse_err = STVParseError(f'invalid vote weight multiplier: {mult!r}'
                              f'on ballot line {line_i}')
    if inner_err is not None:
        raise parse_err from inner_err
    else:
        raise parse_err


def _load_ordered_votes(lines: Iterable[Tuple[Number, List[str]]],
                        nicks: Dict[str, Candidate],
                        ) -> Dict[Tuple[Candidate, ...], Number]:
    votes = collections.defaultdict(int)
    candidates = list(nicks.values())
    for line_i, line_cont in enumerate(lines):
        mult, items = line_cont
        cand_order = []
        for item_i, item in enumerate(items):
            if item.isdigit():
                cand_order.append((candidates[item_i], int(item)))
            elif item != '-':
                raise STVParseError(f'invalid ordered vote item: {item!r}'
                                    f'on ballot line {line_i}')
        cand_order.sort(key=operator.itemgetter(1))
        vote, indices = zip(*cand_order)
        if indices != tuple(range(1, len(indices)+1)):
            raise STVParseError(f'invalid ranking indices: {indices!r}'
                                f' on ballot line {line_i}')
        votes[vote] += mult
    return votes


def _load_unordered_votes(lines: Iterable[Tuple[Number, List[str]]],
                          nicks: Dict[str, Candidate],
                          ) -> Dict[Tuple[Candidate, ...], Number]:
    votes = collections.defaultdict(int)
    for line_i, line_cont in enumerate(lines):
        mult, items = line_cont
        try:
            vote = tuple(nicks[item] for item in items)
        except KeyError as err:
            raise STVParseError(f'unknown candidate in ballot line {line_i}') \
                from err
        votes[vote] += mult
    return votes


def _create_system(title: Optional[str] = None,
                   method: Optional[str] = None,
                   quota: Optional[Union[str, Tuple[str, str]]] = None,
                   seats: Optional[str] = None,
                   random: Optional[str] = None,
                   ) -> votelib.VotingSystem:
    if title is not None and not isinstance(title, str):
        if isinstance(title, tuple):
            raise STVParseError('duplicate election title')
        else:
            raise STVParseError(f'invalid election title: {title!r}')
    return votelib.VotingSystem(title, _create_evaluator(
        method=method, quota=quota, seats=seats, random=random
    ))


def _create_evaluator(method: Optional[str] = None,
                      quota: Optional[Union[str, Tuple[str, str]]] = None,
                      seats: Optional[str] = None,
                      random: Optional[str] = None,
                      ) -> votelib.evaluate.Evaluator:
    mandatory_quota = False
    if method == 'GPCA2000':
        method = 'BC'
        quota = ('droop', 'mandatory')
    if method == 'BC':
        transferer = 'Gregory'
    elif not method:
        raise STVParseError('STV method not found')
    elif method != 'blt':
        raise NotImplementedError(f'STV method not implemented: {method!r}')
    if isinstance(quota, tuple):
        if len(quota) > 2:
            raise STVParseError(f'too many quota settings: {quota!r}')
        elif 'mandatory' in quota:
            mandatory_quota = True
            quota = tuple(item for item in quota if item != 'mandatory')[0]
        else:
            raise STVParseError(f'unknown quota settings: {quota!r}')
    if quota is None:
        if method != 'blt':
            raise STVParseError('quota setting not found')
        # otherwise, we will not need quota_function
    elif quota.isdigit():
        quota_function = votelib.component.quota.constant(int(quota))
    else:
        try:
            quota_function = votelib.component.quota.get(quota)
        except KeyError:
            raise STVParseError(f'unknown quota type: {quota!r}')
    if method == 'blt':
        evaluator = votelib.evaluate.UnknownEvaluator()
    else:
        evaluator = TransferableVoteSelector(
            quota_function=quota_function,
            transferer=transferer,
            mandatory_quota=mandatory_quota,
        )
    if random:
        evaluator = _add_tiebreaker(evaluator, random)
    if seats:
        evaluator = _add_fixed_seats(evaluator, seats)
    return evaluator


def _add_tiebreaker(evaluator: votelib.evaluate.Evaluator,
                    random: str,
                    ) -> votelib.evaluate.Evaluator:
    if random == 'non':
        tiebreaker = votelib.evaluate.auxiliary.CandidateNumberRanker()
    elif random.isdigit():
        tiebreaker = votelib.evaluate.auxiliary.Sortitor(seed=int(random))
    else:
        raise STVParseError('invalid random= parameter: {random!r}')
    return votelib.evaluate.TieBreaking(
        evaluator,
        votelib.evaluate.PreConverted(
            votelib.convert.RankedToPresenceCounts(),
            tiebreaker
        )
    )


def _add_fixed_seats(evaluator: votelib.evaluate.Evaluator,
                     seats: str,
                     ) -> votelib.evaluate.FixedSeatCount:
    try:
        n_seats = int(seats)
    except ValueError as err:
        raise STVParseError(f'invalid seat count: {seats!r}') from err
    return votelib.evaluate.FixedSeatCount(
        evaluator, n_seats=n_seats
    )


load, loads = votelib.io.core.loaders(load_lines)



'''Candidate specifications and nomination validators.

Contains definitions of candidate types and interfaces (:class:`Candidate`,
:class:`ElectionParty`, :class:`Coalition`), constituencies
(:class:`Constituency`), special vote variants (:class:`NoneOfTheAbove`,
:class:`ReopenNominations`) and candidate/nomination validators
(:class:`BasicNominator`, :class:`PersonNominator`, :class:`PartyNominator`).

Apart from the candidate objects specified here, Votelib also accepts strings
with candidate names in all concerned places. The subclasses of this class
are thus only needed for systems considering candidate properties
such as coalition size-dependent electoral thresholds.
For a particular example of this, the
:class:`votelib.evaluate.threshold.PropertyBracketer` chooses different rules
based on a value of an arbitrary property of the candidate. You can use any
object that has the specified property defined; the classes defined here can be
assigned arbitrary additional properties, so they, too, can be used.
'''

from __future__ import annotations

import abc
from typing import Any, List, Optional, Union, Dict

from votelib.persist import simple_serialization


class CandidateError(Exception):
    '''A candidate is invalid in the given context.

    E.g. parties in place of individual candidates, or candidates ineligible
    to stand for given seats.

    :param candidate: Candidate that was found to be invalid.
    :param expected: Definition of a candidate that was expected.
    '''
    def __init__(self, candidate: Any, expected: Any = None):
        self.candidate = candidate
        self.expected = expected
        message = f'invalid candidate: {candidate}'
        if expected:
            message += f', must be {expected}'
        super().__init__(message)


class CandidateObject(metaclass=abc.ABCMeta):
    '''An abstract class for election candidates.

    The subclass check is overridden so that any hashable object that is not
    a set or tuple is accepted. Subclasses will not inherit this override.

    Apart from subclasses of this class, Votelib also accepts strings with
    candidate names in all concerned places. The subclasses of this class
    are thus only needed for systems considering candidate properties
    such as coalition size-dependent electoral thresholds.
    '''

    withdrawn: bool = False
    '''Whether the candidate withdrew from the election (and is thus ineligible
    to get elected).'''


Candidate = Union[str, CandidateObject]


class IndividualElectionOption(CandidateObject):
    '''An abstract class for individuals standing for an election.'''

    candidacy_for: Optional[ElectionParty] = NotImplemented
    '''The party for which this choice is candidating for the election.

    This is important for counting party-based election results where the votes
    are cast and counted for candidates.

    If the candidate is endorsed by multiple parties, use a :class:`Coalition`
    object.
    '''

    membership: Optional[PoliticalParty] = NotImplemented
    '''The party the candidate is member of.

    This is important for counting party-based election results where the votes
    are cast and counted for candidates.
    '''


@simple_serialization
class Person(IndividualElectionOption):
    '''A physical person standing for the election.

    :param name: Name of the person, in any customary text format.
    :param number: Candidacy number assigned to the person for the purpose
        of the election; usually drawn by lot.
    :param membership: A political party the person is member of, if any.
    :param candidacy_for: A political party for which the person is standing
        in the election. If there is no such party, the person is considered
        an independent candidate.
    :param withdrawn: Whether the candidate withdrew from the election
        (and is thus ineligible to get elected).
    '''
    def __init__(self,
                 name: str,
                 number: Optional[int] = None,
                 membership: Optional[PoliticalParty] = None,
                 candidacy_for: Optional[ElectionParty] = None,
                 properties: Dict[str, Any] = None,
                 withdrawn: bool = False,
                 ):
        self.name = name
        self.number = number
        self.membership = membership
        self.candidacy_for = candidacy_for
        self.properties = properties if properties is not None else {}
        self.withdrawn = withdrawn

    def __repr__(self) -> str:
        return (
            f'<Person({self.name}'
            + (f',{self.number}' if self.number is not None else '')
            + ')>'
        )


class ElectionParty(CandidateObject):
    '''A subject that is regarded as a political party for the election.'''

    is_coalition = NotImplemented
    '''Whether the party is a coalition.

    This makes a difference for some systems; the most common case is a
    heightened vote threshold in proportional elections.
    '''


@simple_serialization
class PoliticalParty(ElectionParty):
    '''A political party or movement that is eligible to stand in elections.

    :param name: Name of the party, in any customary text format.
    :param number: Candidacy number assigned to the party for the purpose
        of the election; usually drawn by lot.
    :param affiliation: Other (e.g. national or supranational) parties this
        party is affiliated with, if any.
    :param lead: A person that leads the party into the elections.
    '''
    is_coalition = False

    def __init__(self,
                 name: str,
                 number: Optional[int] = None,
                 affiliations: Optional[List[PoliticalParty]] = None,
                 lead: Optional[Person] = None,
                 properties: Dict[str, Any] = None,
                 withdrawn: bool = False,
                 ):
        self.name = name
        self.number = number
        self.affiliations = affiliations
        self.lead = lead
        self.properties = properties if properties is not None else {}
        self.withdrawn = withdrawn

    def __repr__(self) -> str:
        return (
            f'<PoliticalParty({self.name}'
            + (f',{self.number}' if self.number is not None else '')
            + ')>'
        )


@simple_serialization
class Coalition(ElectionParty):
    '''A coalition of two or more election-eligible parties.

    :param parties: Parties involved in the coalition.
    :param name: Name of the coalition, in any customary text format. If not
        given, it is derived from the names of the member parties.
    :param number: Candidacy number assigned to the coalition for the purpose
        of the election; usually drawn by lot.
    :param affiliations: Other (e.g. national or supranational) parties this
        coalition is affiliated with, if any.
    :param lead: A person that leads the coalition into the elections.
    '''
    is_coalition = True

    def __init__(self,
                 parties: List[PoliticalParty],
                 name: Optional[str] = None,
                 number: Optional[int] = None,
                 affiliations: Optional[List[PoliticalParty]] = None,
                 lead: Optional[Person] = None,
                 withdrawn: bool = False,
                 ):
        self.parties = parties
        if name is None:
            name = '-'.join(p.name for p in parties)
        self.name = name
        self.number = number
        self.affiliations = affiliations
        self.lead = lead
        self.withdrawn = withdrawn

    def get_n_coalition_members(self) -> int:
        '''Return the number of member parties in the coalition.

        This is important in some elections which specify different thresholds
        for coalitions with a given number of members.
        '''
        return len(self.parties)

    def __repr__(self) -> str:
        return (
            'Coalition('
            + repr(self.parties)
            + (f',{self.number}' if self.number is not None else '')
            + ')'
        )


@simple_serialization
class IndividualToPartyMapper:
    '''Define mapping from individuals to parties.

    A candidate object such as a :class:`Person` might define multiple party
    relationships or define that the candidate is standing for the election as
    an independent. This object can be added as a component to some converters
    or evaluators to map individuals to parties correctly according to the
    rules of the particular election system.

    :param affiliation: How to allocate the candidate to the party:

        -   `'candidacy_for'` (default) uses the party the candidate stood
            for (was endorsed by) in the election,
        -   `'membership'` uses the party the candidate is a member of.

    :param independents: How to handle individual candidates not candidating
        for a party:

        -   `'aggregate'` (default) aggregates them under the None key
            to a total count,
        -   `'keep'` keeps them separately,
        -   `'ignore'` omits them from the result,
        -   `'error'` raises an error.
    '''

    AFFILIATION_METHODS: List[str] = [
        'candidacy_for', 'membership'
    ]
    INDEPENDENTS_METHODS: List[str] = [
        'error', 'keep', 'aggregate', 'ignore'
    ]

    IGNORE = object()

    def __init__(self,
                 affiliation: str = 'candidacy_for',
                 independents: str = 'aggregate',
                 ):
        if affiliation not in self.AFFILIATION_METHODS:
            raise ValueError(
                'invalid assignment of individuals to parties:'
                f' {affiliation}, allowed {self.AFFILIATION_METHODS}'
            )
        if independents not in self.INDEPENDENTS_METHODS:
            raise ValueError(
                'invalid handling of independents in aggregation:'
                f' {independents}, allowed {self.INDEPENDENTS_METHODS}'
            )
        self.affiliation = affiliation
        self.independents = independents

    def __call__(self,
                 cand: IndividualElectionOption
                 ) -> Union[ElectionParty, IndividualElectionOption, None]:
        party = getattr(cand, self.affiliation)
        if party is None:
            if self.independents == 'error':
                raise CandidateError(
                    f'independent {cand} not allowed in aggregation'
                )
            elif self.independents == 'keep':
                return cand
            elif self.independents == 'aggregate':
                return None
            else:
                return self.IGNORE
        else:
            return party


@simple_serialization
class BlankVoteOption(ElectionParty, IndividualElectionOption):
    '''A base class for blank (non-partisan) vote choices.

    This includes votes that are not counted to any candidate. In some rare
    cases, these votes have a special effect (such as triggering a new election
    if there is a sufficient number of them), but most of the time they can be
    safely disregarded.
    '''
    is_coalition = False
    candidacy_for = None
    membership = None

    def __init__(self,
                 name: str,
                 ):
        self.name = name


class NoneOfTheAbove(BlankVoteOption):
    '''None of the above (NOTA) vote option (often called white ballots).

    In some contexts, a sufficient number of these votes can trigger a special
    effect such as a new election.

    :param name: Actual name of the NOTA option in the given context.
    '''


class ReopenNominations(BlankVoteOption):
    '''Reopen Nominations vote option.

    In some elections, this option is present; if it prevails, it can trigger
    a new round of nominations and a new election.

    :param name: Actual name of the Reopen Nominations option in the given
        context.
    '''


class Constituency:
    '''A constituency for elections with multiple sets of candidates.

    Constituencies are used where the electorate is separated into more than
    one group. The most common variant is spatial - electoral or voting
    districts (also called wards or precincts). This object should be used
    only for cases where such spatial division is used to evaluate the result,
    not merely used to collect and count the ballots.

    Non-spatial variants include
    curias (used in 19th century Austria-Hungary for different social classes),
    ethnical divisions (Maori electorates in New Zealand)
    or qualificational divisions (University constituencies in Ireland).

    In the current version of Votelib, special constituency objects are not
    needed (any hashable objects such as strings can be used for all
    functionality implemented so far and will continue to be supported by the
    existing features in the future) so this is essentially just a type marker,
    but further development might necessitate creating some subclasses of this,
    and more specific interfaces.
    '''


class Nominator(metaclass=abc.ABCMeta):
    '''An abstract class for nominators (candidacy validators).'''
    @abc.abstractmethod
    def validate(self, candidate: Candidate) -> None:
        '''Check if the candidate satisfies criteria given by the system.

        :raises NotImplementedError:
        '''
        raise NotImplementedError


# Nominators (checkers)
@simple_serialization
class BasicNominator(Nominator):
    '''Validate that the election candidates are valid objects.

    Does not do any logical checks; only validates that the candidate instance
    passes the criteria of the :class:`Candidate` class (such as checking
    against most types of collections).

    :param allow_blank: Whether to allow blank votes (NOTA, ReopenNominations).
    '''
    def __init__(self, allow_blank: bool = True):
        self.allow_blank = allow_blank

    def validate(self, candidate: Candidate) -> None:
        '''Check whether a candidate is valid.

        :param candidate: Candidate to be checked.
        :raises CandidateError: If a candidate is invalid.
        '''
        if not isinstance(candidate, (str, CandidateObject)):
            raise CandidateError(f'invalid candidate {candidate}')
        if not self.allow_blank and isinstance(candidate, BlankVoteOption):
            raise CandidateError('blank vote')


@simple_serialization
class PersonNominator(Nominator):
    '''Validate that election candidates are physical persons and not parties.

    :param allow_independents: Whether persons candidating without a support
        of a political party can stand in the election.
    :param allow_blank: Whether to allow blank votes (NOTA, ReopenNominations).
    '''
    def __init__(self,
                 allow_independents: bool = True,
                 allow_blank: bool = True,
                 ):
        self.allow_independents = allow_independents
        self.allow_blank = allow_blank

    def validate(self, candidate: Candidate) -> None:
        '''Check whether a candidate is valid.

        :param candidate: Candidate to be checked.
        :raises CandidateError: If a candidate is invalid.
        '''
        if not isinstance(candidate, IndividualElectionOption):
            raise CandidateError(candidate, IndividualElectionOption)
        if isinstance(candidate, BlankVoteOption):
            if not self.allow_blank:
                raise CandidateError(candidate, 'non-blank vote')
        elif not self.allow_independents and not candidate.candidacy_for:
            raise CandidateError(
                candidate, 'a candidate with party candidacy support'
            )


@simple_serialization
class PartyNominator(Nominator):
    '''Check that election candidates are electoral parties, not persons.

    This includes political parties in the broadest sense of the term, as well
    as their coalitions formed for the purpose of the election.

    Only instances of :class:`ElectionParty` and its subclasses are accepted,
    as are classes that pass the interface check of the class, if there is one.
    Blank votes are accepted by default and can be disallowed, like coalitions.

    :param allow_coalitions: Whether to allow coalitions.
    :param allow_blank: Whether to allow blank votes (NOTA, ReopenNominations).
    '''
    def __init__(self,
                 allow_coalitions: bool = True,
                 allow_blank: bool = True
                 ):
        self.allow_coalitions = allow_coalitions
        self.allow_blank = allow_blank

    def validate(self, candidate: Candidate) -> None:
        '''Check whether a candidate is a valid election party.

        :param candidate: Candidate to be checked.
        :raises CandidateError: If a candidate is not a valid election party
            candidate.
        '''
        if isinstance(candidate, BlankVoteOption):
            if not self.allow_blank:
                raise CandidateError(candidate, 'non-blank vote')
        elif not isinstance(candidate, ElectionParty):
            raise CandidateError(candidate, ElectionParty)
        elif not self.allow_coalitions and isinstance(candidate, Coalition):
            raise CandidateError(candidate, 'non-coalition')



'''Cardinal voting systems - systems that use score votes.

These systems have the most complicated input - each voter can assign a range
of scores to candidates - but are claimed to circumvent the Arrow's
impossibility theorem and Gibbard-Satterthwaite theorem (but not the more
general Gibbard's theorem) that only hold generally for ordinal (ranked) voting
systems.
'''

import math
from fractions import Fraction
from typing import Any, List, Dict, Union, Callable
from numbers import Number

import votelib.convert
import votelib.persist
import votelib.evaluate.core
import votelib.evaluate.condorcet
from votelib.candidate import Candidate
from votelib.vote import ScoreVoteType


class ScoreVoting:
    '''Evaluate ordinary score voting (range voting) systems.

    With the aggregation function set to sum, it also evaluates cumulative
    voting systems.

    This is essentially just a :class:`votelib.convert.ScoreToSimpleVotes`
    (to which all parameters are passed to) followed by a
    :class:`votelib.evaluate.core.Plurality` evaluator that selects the highest
    aggregate score.
    '''
    def __init__(self,
                 function: Union[
                     Callable[[List[Number]], Number], str
                 ] = 'mean',
                 unscored_value: Union[
                     Callable[[List[Number]], Number], str, Number, None
                 ] = None,
                 min_count: int = 0,
                 truncation: Number = 0,
                 bottom_value: Number = 0,
                 ):
        self._agg = votelib.convert.ScoreToSimpleVotes(
            function=function,
            unscored_value=unscored_value,
            min_count=min_count,
            truncation=truncation,
            bottom_value=bottom_value,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            **self._agg.to_dict(),
            'class': votelib.persist.scoped_class_name(self)
        }

    def evaluate(self,
                 votes: Dict[ScoreVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by highest score.

        :param votes: Score votes.
        :param n_seats: Number of candidates to select.
        '''
        return votelib.evaluate.core.get_n_best(
            self._agg.convert(votes),
            n_seats
        )


class MajorityJudgment:
    '''Majority Judgment, a median-based cardinal voting system.

    Parameters other than tie_breaking are passed to
    :class:`votelib.convert.ScoreToSimpleVotes`.

    Note: This evaluator does not return elected candidates in order of
    precedence (it does not resolve their exact ordering).

    :param tie_breaking: Method of breaking ties for candidates with equal
        median ranking. The `'default'` method (original by Balinski) removes
        median votes until a group of N winners has a higher ranking
        (known to be susceptible to electing candidates with consolidated bases
        rather than candidates with broad support),
        the `'plus'` method by Bosworth selects N candidates with the greatest
        number of scores equal or higher than the median (which claims to
        address the above shortcoming).
    :param unscored_value: Score to give to a candidate that was not assigned
        a score by the voter. None means such ballots will not be considered
        for the candidate. A function can be specified to determine the value
        from all assigned scores (such as assigning the minimum). Builtin
        functions and functions from the statistics stdlib module can be
        specified by their names.
    :param min_count: Minimum count of voter scores for the candidate to be
        considered. Candidates below this threshold will be assigned
        bottom_value.
    :param truncation: Fraction (if lower than 1) or count (if at least 1)
        of lowest and highest scores to disregard before aggregating, to
        stabilize the result. (Both ends are trimmed using this, so the
        number/fraction of scores disregarded is twice the count/fraction.)
    :param bottom_value: Value to assign to candidates with less voter scores
        than min_count. This would usually be the lowest possible aggregate
        score.
    '''

    def __init__(self,
                 tie_breaking: str = 'default',
                 unscored_value: Union[
                     Callable[[List[Number]], Number], str, Number, None
                 ] = None,
                 min_count: int = 0,
                 truncation: Number = 0,
                 bottom_value: Number = 0,
                 ):
        self._agg = votelib.convert.ScoreToSimpleVotes(
            function='median_low',
            unscored_value=unscored_value,
            min_count=min_count,
            truncation=truncation,
            bottom_value=bottom_value,
        )
        self.tie_breaking = tie_breaking
        self.tie_breaker = getattr(self, '_tiebreak_' + tie_breaking)

    def to_dict(self) -> Dict[str, Any]:
        out = {
            'class': votelib.persist.scoped_class_name(self),
            'tie_breaking': self.tie_breaking,
        }
        for key, val in self._agg.to_dict().items():
            if key not in ('function', 'class'):
                out[key] = val
        return out

    def evaluate(self,
                 votes: Dict[Candidate, Dict[Number, int]],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by Majority Judgment (appx. highest median score).

        :param votes: Score votes.
        :param n_seats: Number of candidates to select.
        '''
        corrected_scores = self._agg.corrected_scores(votes)
        order = votelib.evaluate.core.get_n_best(
            self._agg.aggregate(corrected_scores),
            n_seats
        )
        if isinstance(order[-1], votelib.evaluate.core.Tie):
            n_tied_seats = order.count(order[-1])
            return order[:-n_tied_seats] + self.tie_breaker({
                cand: corrected_scores[cand] for cand in order[-1]
            }, n_tied_seats)
        else:
            return order

    def _tiebreak_default(self,
                          scores: Dict[Candidate, Dict[Any, int]],
                          n_seats: int,
                          ) -> List[Candidate]:
        '''Perform Majority Judgment tiebreaking by removing median scores.

        This removes median scores from tied candidates until the median of the
        remaining scores differs among them, and then selects the one with the
        highest new median score.
        '''
        scores = {cand: cscores.copy() for cand, cscores in scores.items()}
        while max(sum(cscores.values()) for cscores in scores.values()):
            medians = self._agg.aggregate(scores)
            best = votelib.evaluate.core.get_n_best(medians, n_seats)
            for i, result in enumerate(best):
                if isinstance(result, votelib.evaluate.core.Tie):
                    if i > 0:
                        winners = best[:i]
                        return winners + self._tiebreak_default({
                            cand: scores[cand] for cand in scores.keys()
                            if cand not in winners
                        }, n_seats - i)
                    else:
                        break
            else:
                return best
            closest_change = self._closest_median_change(scores, medians)
            if closest_change == 0:
                closest_change = 1
            for cand, cscores in scores.items():
                cscores[medians[cand]] -= closest_change
        raise votelib.evaluate.core.VotingSystemError(
            'cannot determine clear cutoff'
        )

    @staticmethod
    def _closest_median_change(scores: Dict[Candidate, Dict[Any, int]],
                               medians: Dict[Candidate, Any],
                               ) -> int:
        closest_change = float('inf')
        for cand, cscores in scores.items():
            lower, upper = 0, 0
            half = Fraction(sum(cscores.values()), 2)
            cur_median = medians[cand]
            for score, count in cscores.items():
                if score >= cur_median:
                    lower += count
                if score > cur_median:
                    upper += count
            cand_closest = int(min(
                math.ceil(abs(lower - half)),
                math.ceil(abs(upper - half))
            ))
            if cand_closest < closest_change:
                closest_change = cand_closest
        return closest_change

    def _tiebreak_plus(self,
                       scores: Dict[Candidate, Dict[Number, int]],
                       n_seats: int,
                       ) -> List[Candidate]:
        '''Perform Majority Judgment Plus tiebreaking.

        This takes the candidates with the highest amount of scores higher or
        equal to the median, as suggested by Bosworth.
        '''
        # select the one with largest majority with the largest median grade
        # so we need to find out the median grade again
        majorities = self._counts_over_score(
            scores, self._shared_median(scores), strict=False
        )
        return votelib.evaluate.core.get_n_best(majorities, n_seats)

    def _shared_median(self,
                       scores: Dict[Candidate, Dict[Number, int]]
                       ) -> Number:
        return self._agg.aggregate_one(next(iter(scores.values())))

    def _counts_over_score(self,
                           scores: Dict[Candidate, Dict[Number, int]],
                           threshold: Number,
                           strict: bool = False
                           ) -> Dict[Candidate, int]:
        return {
            cand: sum(
                count for score, count in cscores.items()
                if score > threshold or (not strict and score == threshold)
            ) for cand, cscores in scores.items()
        }


class STAR:
    '''Score Then Automatic Run-Off (STAR) cardinal voting system.

    A score-based voting system that aims to reduce suceptibility to tactical
    voting by forcing a run-off between the highest-ranked candidates.

    STAR voting is normally specified for single-winner elections only, with
    two candidates in the run-off, but here it is extended:

    -   To enable multi-winner selections (n_seats more than 1), the run-off is
        performed using a selected Condorcet evaluator based on pairwise score
        preferences. This gives an identical result to simple plurality
        evaluation with only two candidates in the run-off, but also creates a
        robust result when there are more.
    -   Finer control over the run-off size is allowed; you can specify more
        than one extra candidate in the run-off, or specify the run-off size
        as a fraction of the number of seats to be filled.

    For single-winner elections, keep the runoff_added_* parameters at their
    defaults; then, the choice of runoff_evaluator does not matter because all
    Condorcet evaluators behave the same for two-candidate elections.

    :param runoff_added_count: Number of extra candidates in the run-off (added
        to the number of seats to be filled) as an absolute number.
    :param runoff_added_fraction: Number of extra candidates in the run-off
        (added to the number of seats to be filled) as a fraction of the number
        of seats to be filled.
    :param runoff_evaluator: Name of the Condorcet evaluator to use for the
        run-off (references `votelib.evaluate.condorcet.EVALUATORS`),
        or an instance of such a selector (must accept votes in pairwise win
        format).
    :param unscored_value: Score to give to a candidate that was not assigned
        a score by the voter. None means such ballots will not be considered
        for the candidate. A callable is not accepted.
    :param min_count: Minimum count of voter scores for the candidate to be
        considered. Candidates below this threshold will be assigned
        bottom_value.
    :param truncation: Fraction (if lower than 1) or count (if at least 1)
        of lowest and highest scores to disregard before aggregating, to
        stabilize the result. (Both ends are trimmed using this, so the
        number/fraction of scores disregarded is twice the count/fraction.)
    :param bottom_value: Value to assign to candidates with less voter scores
        than min_count. This would usually be the lowest possible aggregate
        score.
    '''
    def __init__(self,
                 runoff_added_count: int = 1,
                 runoff_added_fraction: Number = 0,
                 runoff_evaluator: Union[
                     str,
                     votelib.evaluate.condorcet.Selector
                 ] = 'schulze',
                 unscored_value: Union[str, Number, None] = None,
                 min_count: int = 0,
                 truncation: Number = 0,
                 bottom_value: Number = 0,
                 ):
        if isinstance(runoff_evaluator, str):
            runoff_evaluator = votelib.evaluate.condorcet.EVALUATORS[
                runoff_evaluator
            ]
        self._agg = votelib.convert.ScoreToSimpleVotes(
            function='sum',
            unscored_value=unscored_value,
            min_count=min_count,
            truncation=truncation,
            bottom_value=bottom_value,
        )
        self._runoff_torank_conv = votelib.convert.ScoreToRankedVotes(
            unscored_value=unscored_value
        )
        self._runoff_tocond_conv = votelib.convert.RankedToCondorcetVotes()
        self.runoff_added_count = runoff_added_count
        self.runoff_added_fraction = runoff_added_fraction
        self.runoff_evaluator = runoff_evaluator

    def to_dict(self) -> Dict[str, Any]:
        out = {
            'class': votelib.persist.scoped_class_name(self),
            'runoff_added_count': self.runoff_added_count,
            'runoff_added_fraction': self.runoff_added_fraction,
            'runoff_evaluator': self.runoff_evaluator.to_dict(),
        }
        for key, val in self._agg.to_dict().items():
            if key not in ('function', 'class'):
                out[key] = val
        return out

    def evaluate(self,
                 votes: Dict[ScoreVoteType, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by STAR voting.

        :param votes: Score votes.
        :param n_seats: Number of candidates to select.
        '''
        # Aggregate scores to determine who goes to runoff.
        agg_scores = self._agg.convert(votes)
        runoff_size = (
            n_seats + self.runoff_added_count
            + int(math.ceil(self.runoff_added_fraction * n_seats))
        )
        runoff_members = votelib.evaluate.core.get_n_best(
            agg_scores, runoff_size
        )
        # Convert scores to pairwise wins for the runoff.
        pairwin_votes = {
            pair: n_votes
            for pair, n_votes in self._runoff_tocond_conv.convert(
                self._runoff_torank_conv.convert(votes)
            ).items()
            # Only preferences for runoff members.
            if all(cand in runoff_members for cand in pair)
        }
        return self.runoff_evaluator.evaluate(pairwin_votes, n_seats)



import setuptools

with open('README.md') as infile:
    long_description = infile.read()

with open('VERSION') as infile:
    version = infile.read().strip()

setuptools.setup(
    name='votelib',
    version=version,
    description='Voting evaluation library for Python',
    long_description=long_description,
    long_description_content_type='text/markdown; charset=UTF-8',
    author='Jan Šimbera',
    author_email='simbera.jan@gmail.com',
    python_requires='>=3.7.0',
    url='https://github.com/simberaj/votelib',
    packages=setuptools.find_packages(exclude=('tests', )),
    install_requires=[],
    extras_require={},
    include_package_data=True,
    license='MIT',
    keywords='voting election vote electoral apportionment condorcet python',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    zip_safe=True
)
